from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sqlite3
import csv

import ga4gh.server.exceptions as exceptions


SUPPORTED_RNA_INPUT_FORMATS = ["cufflinks", "kallisto", "rsem"]


class RnaSqliteStore(object):
    """
    Defines a sqlite store for RNA data as well as methods for loading the
    tables.
    """
    def __init__(self, sqliteFileName):
        self._dbConn = sqlite3.connect(sqliteFileName)
        self._cursor = self._dbConn.cursor()
        self._batchSize = 2000
        self._rnaValueList = []
        self._expressionValueList = []

    def createTables(self):
        # annotationIds is a comma separated list
        self._cursor.execute('''CREATE TABLE RnaQuantification (
                       id TEXT NOT NULL PRIMARY KEY,
                       feature_set_ids TEXT,
                       description TEXT,
                       name TEXT,
                       read_group_ids TEXT,
                       programs TEXT,
                       biosample_id TEXT)''')
        self._cursor.execute('''CREATE TABLE Expression (
                       id INTEGER,
                       rna_quantification_id TEXT,
                       name TEXT,
                       expression REAL,
                       is_normalized BOOLEAN,
                       raw_read_count REAL,
                       score REAL,
                       units INTEGER,
                       conf_low REAL,
                       conf_hi REAL,
                       PRIMARY KEY (id, rna_quantification_Id))''')
        self._dbConn.commit()

    def addRNAQuantification(self, datafields):
        """
        Adds an RNAQuantification to the db.  Datafields is a tuple in the
        order:
        id, feature_set_ids, description, name,
        read_group_ids, programs, biosample_id
        """
        self._rnaValueList.append(datafields)
        if len(self._rnaValueList) >= self._batchSize:
            self.batchaddRNAQuantification()

    def batchaddRNAQuantification(self):
        if len(self._rnaValueList) > 0:
            sql = "INSERT INTO RnaQuantification VALUES (?,?,?,?,?,?,?)"
            self._cursor.executemany(sql, self._rnaValueList)
            self._dbConn.commit()
            self._rnaValueList = []

    def addExpression(self, datafields):
        """
        Adds an Expression to the db.  Datafields is a tuple in the order:
        id, rna_quantification_id, name, expression,
        is_normalized, raw_read_count, score, units, conf_low, conf_hi
        """
        self._expressionValueList.append(datafields)
        if len(self._expressionValueList) >= self._batchSize:
            self.batchAddExpression()

    def batchAddExpression(self):
        if len(self._expressionValueList) > 0:
            sql = "INSERT INTO Expression VALUES (?,?,?,?,?,?,?,?,?,?)"
            self._cursor.executemany(sql, self._expressionValueList)
            self._dbConn.commit()
            self._expressionValueList = []

    def createIndices(self):
        """
        Index columns that are queried. The expression index can
        take a long time.
        """

        sql = '''CREATE INDEX name_index
                 ON Expression (name)'''
        self._cursor.execute(sql)
        self._dbConn.commit()

        sql = '''CREATE INDEX expression_index
                 ON Expression (expression)'''
        self._cursor.execute(sql)
        self._dbConn.commit()


class AbstractWriter(object):
    """
    Base class to use for the rna quantification writers
    """
    def __init__(self, rnaDB, featureType="gene", dataset=None):
        self._db = rnaDB
        self._isNormalized = None
        self._units = 0  # EXPRESSION_UNIT_UNSPECIFIED
        self._expressionLevelCol = None
        self._idCol = None
        self._nameCol = None
        self._countCol = None
        self._confColLow = None
        self._confColHi = None
        self._dataRepo = None
        self._dataset = dataset
        self._featureType = featureType

    def setUnits(self, units):
        if units == "fpkm":
            self._units = 1
        elif units == "tpm":
            self._units = 2

    def setColNum(self, header, name, defaultNum=None):
        colNum = defaultNum
        try:
            colNum = header.index(name)
        except:
            if defaultNum is None:
                raise exceptions.RepoManagerException(
                       "Missing {} column in expression table.".format(name))
        return colNum

    def writeExpression(self, rnaQuantificationId, quantfilename):
        """
        Reads the quantification results file and adds entries to the
        specified database.
        """
        isNormalized = self._isNormalized
        units = self._units
        with open(quantfilename, "r") as quantFile:
            quantificationReader = csv.reader(quantFile, delimiter=b"\t")
            header = next(quantificationReader)
            expressionLevelColNum = self.setColNum(
                                        header, self._expressionLevelCol)
            nameColNum = self.setColNum(header, self._nameCol)
            countColNum = self.setColNum(header, self._countCol, -1)
            confColLowNum = self.setColNum(header, self._confColLow, -1)
            confColHiNum = self.setColNum(header, self._confColHi, -1)
            expressionId = 0
            for expression in quantificationReader:
                expressionLevel = expression[expressionLevelColNum]
                name = expression[nameColNum]
                rawCount = 0.0
                if countColNum != -1:
                    rawCount = expression[countColNum]
                confidenceLow = 0.0
                confidenceHi = 0.0
                score = 0.0
                if confColLowNum != -1 and confColHiNum != -1:
                    confidenceLow = float(expression[confColLowNum])
                    confidenceHi = float(expression[confColHiNum])
                    score = (confidenceLow + confidenceHi)/2

                datafields = (expressionId, rnaQuantificationId, name,
                              expressionLevel, isNormalized, rawCount, score,
                              units, confidenceLow, confidenceHi)
                self._db.addExpression(datafields)
                expressionId += 1
            self._db.batchAddExpression()


class CufflinksWriter(AbstractWriter):
    """
    Class to parse and write expression data from an input file generated by
    Cufflinks.

    cufflinks header:
        tracking_id    class_code    nearest_ref_id    gene_id
        gene_short_name    tss_id    locus    length    coverage    FPKM
        FPKM_conf_lo    FPKM_conf_hi    FPKM_status
    """
    def __init__(self, rnaDB, featureType, units="fpkm", dataset=None):
        super(CufflinksWriter, self).__init__(
            rnaDB, featureType, dataset=dataset)
        self._isNormalized = True
        self._expressionLevelCol = "FPKM"
        self._idCol = "tracking_id"
        self._nameCol = "tracking_id"
        self._confColLow = "FPKM_conf_lo"
        self._confColHi = "FPKM_conf_hi"
        self.setUnits(units)


class RsemWriter(AbstractWriter):
    """
    Class to parse and write expression data from an input file generated by
    RSEM.

    RSEM header (gene quantification):
    gene_id    transcript_id(s)    length    effective_length    expected_count
    TPM    FPKM    posterior_mean_count
    posterior_standard_deviation_of_count    pme_TPM    pme_FPKM
    TPM_ci_lower_bound    TPM_ci_upper_bound    FPKM_ci_lower_bound
    FPKM_ci_upper_bound

    RSEM header (transcript quantification):
    transcript_id    gene_id    length    effective_length    expected_count
    TPM    FPKM    IsoPct    posterior_mean_count
    posterior_standard_deviation_of_count    pme_TPM    pme_FPKM
    IsoPct_from_pme_TPM    TPM_ci_lower_bound    TPM_ci_upper_bound
    FPKM_ci_lower_bound    FPKM_ci_upper_bound
    """
    def __init__(self, rnaDB, featureType, units="tpm", dataset=None):
        super(RsemWriter, self).__init__(
            rnaDB, featureType=featureType, dataset=dataset)
        self._isNormalized = True
        self._expressionLevelCol = "TPM"
        self._confColLow = "TPM_ci_lower_bound"
        self._confColHi = "TPM_ci_upper_bound"
        self._countCol = "expected_count"
        if self._featureType is "transcript":
            self._idCol = "transcript_id"
        else:
            self._idCol = "gene_id"
        self._nameCol = self._idCol
        self.setUnits(units)


class KallistoWriter(AbstractWriter):
    """
    Class to parse and write expression data from an input file generated by
    RSEM.

    kallisto header:
        target_id    length    eff_length    est_counts    tpm
    """
    def __init__(self, rnaDB, featureType, units="tpm", dataset=None):
        super(KallistoWriter, self).__init__(
            rnaDB, featureType, dataset=dataset)
        self._isNormalized = True
        self._expressionLevelCol = "tpm"
        self._idCol = "target_id"
        self._nameCol = "target_id"
        self._countCol = "est_counts"
        self.setUnits(units)


def writeRnaseqTable(rnaDB, analysisIds, description, annotationId,
                     readGroupId="", programs="", biosampleId=""):
    if readGroupId is None:
        readGroupId = ""
    for analysisId in analysisIds:
        datafields = (analysisId, annotationId, description, analysisId,
                      readGroupId, programs, biosampleId)
        rnaDB.addRNAQuantification(datafields)
    rnaDB.batchaddRNAQuantification()


def writeExpressionTable(writer, data):
    for rnaQuantId, quantFilename in data:
        writer.writeExpression(rnaQuantId, quantFilename)


def rnaseq2ga(quantificationFilename, sqlFilename, localName, rnaType,
              dataset=None, featureType="gene",
              description="", programs="", featureSetNames="",
              readGroupSetNames="", biosampleId=""):
    """
    Reads RNA Quantification data in one of several formats and stores the data
    in a sqlite database for use by the GA4GH reference server.

    Supports the following quantification output types:
    Cufflinks, kallisto, RSEM.
    """
    readGroupSetName = ""
    if readGroupSetNames:
        readGroupSetName = readGroupSetNames.strip().split(",")[0]
    featureSetIds = ""
    readGroupIds = ""
    if dataset:
        featureSetIdList = []
        if featureSetNames:
            for annotationName in featureSetNames.split(","):
                featureSet = dataset.getFeatureSetByName(annotationName)
                featureSetIdList.append(featureSet.getId())
            featureSetIds = ",".join(featureSetIdList)
        # TODO: multiple readGroupSets
        if readGroupSetName:
            readGroupSet = dataset.getReadGroupSetByName(readGroupSetName)
            readGroupIds = ",".join(
                [x.getId() for x in readGroupSet.getReadGroups()])
    if rnaType not in SUPPORTED_RNA_INPUT_FORMATS:
        raise exceptions.UnsupportedFormatException(rnaType)
    rnaDB = RnaSqliteStore(sqlFilename)
    if rnaType == "cufflinks":
        writer = CufflinksWriter(rnaDB, featureType, dataset=dataset)
    elif rnaType == "kallisto":
        writer = KallistoWriter(rnaDB, featureType, dataset=dataset)
    elif rnaType == "rsem":
        writer = RsemWriter(rnaDB, featureType, dataset=dataset)
    writeRnaseqTable(rnaDB, [localName], description, featureSetIds,
                     readGroupId=readGroupIds, programs=programs,
                     biosampleId=biosampleId)
    writeExpressionTable(writer, [(localName, quantificationFilename)])



"""
Tests the file handle cache
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import shutil
import tempfile
import unittest
import uuid

import ga4gh.server.datamodel as datamodel


class TestFileHandleCache(datamodel.PysamFileHandleCache, unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestFileHandleCache, self).__init__()
        unittest.TestCase.__init__(self, *args, **kwargs)

    def setUp(self):
        self._tempdir = tempfile.mkdtemp(prefix="ga4gh_file_cache",
                                         dir=tempfile.gettempdir())

    def _getFileHandle(self, dataFile):
        def openMethod(dataFile):
            return open(dataFile, 'w')
        return self.getFileHandle(dataFile, openMethod)

    def testGetFileHandle(self):
        def genFileName(x):
            return os.path.join(self._tempdir, str(uuid.uuid4()))

        # Set cache size to 9 files max
        self.setMaxCacheSize(9)

        # Build a list of 10 files and add their handles to the cache
        fileList = map(genFileName, range(0, 10))

        for f in fileList:
            handle = self._getFileHandle(f)
            self.assertEquals(self._cache.count((f, handle)), 1)

        self.assertEquals(len(self._memoTable), len(self._cache))

        # Ensure that the first added file has been removed from the cache
        self.assertEquals(filter(lambda x: x[0] == fileList[0], self._cache),
                          [])

        topIndex = len(self._cache) - 1

        # Update priority of this file and ensure it's no longer the
        # least recently used
        self.assertEquals(self._cache[topIndex][0], fileList[1])
        self._getFileHandle(fileList[1])
        self.assertNotEqual(self._cache[topIndex][0], fileList[1])
        self.assertEquals(self._cache[0][0], fileList[1])

    def testSetCacheMaxSize(self):
        self.assertRaises(ValueError, self.setMaxCacheSize, 0)
        self.assertRaises(ValueError, self.setMaxCacheSize, -1)

    def tearDown(self):
        shutil.rmtree(self._tempdir)



import json
import logging

from yubico_client import yubico_exceptions
from yubico_client.yubico import Yubico
from oic.utils.http_util import Response
from provider.authn import AuthnModule, make_cls_from_name

logger = logging.getLogger(__name__)


class YubicoOTP(AuthnModule):
    url_endpoint = "/yubi_otp/verify"

    def __init__(
            self, yubikey_db, validation_server, client_id, template_env,
            secret_key=None, verify_ssl=True, template="yubico_otp.jinja2",
            **kwargs):
        super(YubicoOTP, self).__init__(None)
        self.template_env = template_env
        self.template = template

        cls = make_cls_from_name(yubikey_db["class"])
        self.yubikey_db = cls(**yubikey_db["kwargs"])

        self.client = Yubico(client_id, secret_key,
                             api_urls=[validation_server],
                             verify_cert=verify_ssl)
        if not verify_ssl:
            # patch yubico-client to not find any ca bundle
            self.client._get_ca_bundle_path = lambda: None

    def __call__(self, *args, **kwargs):
        template = self.template_env.get_template(self.template)
        return Response(template.render(action=self.url_endpoint,
                                        state=json.dumps(kwargs)))

    def verify(self, *args, **kwargs):
        otp = kwargs["otp"]
        try:
            status = self.client.verify(otp, return_response=True)
        except yubico_exceptions.InvalidClientIdError as e:
            logger.error(
                "Client with id {} does not exist".format(e.client_id))
            return self.FAILED_AUTHN
        except yubico_exceptions.SignatureVerificationError:
            logger.error("Signature verification failed")
            return self.FAILED_AUTHN
        except yubico_exceptions.StatusCodeError as e:
            logger.error(
                "Negative status code was returned: {}".format(
                    e.status_code))
            return self.FAILED_AUTHN

        if status:
            logger.debug("Success, the provided OTP is valid")
            yubikey_public_id = otp[:12]

            return self.yubikey_db[yubikey_public_id], True
        else:
            logger.error(
                "No response from the servers "
                "or received other negative status code")



"""
Tests related to exceptions
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import mock
import unittest
import inspect

import ga4gh.server.exceptions as exceptions
import ga4gh.server.frontend as frontend

import ga4gh.schemas.protocol as protocol


class TestExceptionHandler(unittest.TestCase):
    """
    Test that caught exceptions are handled correctly
    """
    @classmethod
    def setUpClass(cls):
        frontend.reset()
        frontend.configure(baseConfig="TestConfig")
        frontend.app.log_exception = mock.Mock()

    class UnknownException(Exception):
        pass

    def getGa4ghException(self, data):
        return protocol.fromJson(data, protocol.GAException)

    def testObjectNotFoundException(self):
        exception = exceptions.ObjectNotFoundException()
        response = frontend.handleException(exception)
        self.assertEquals(response.status_code, 404)
        self.assertFalse(frontend.app.log_exception.called)

    def testCallSetNotInVariantSetException(self):
        exception = exceptions.CallSetNotInVariantSetException(
            'csId', 'vsId')
        response = frontend.handleException(exception)
        self.assertEquals(response.status_code, 404)
        gaException = self.getGa4ghException(response.data)
        self.assertGreater(len(gaException.message), 0)
        self.assertFalse(frontend.app.log_exception.called)

    def testUnknownExceptionBecomesServerError(self):
        exception = self.UnknownException()
        response = frontend.handleException(exception)
        self.assertEquals(response.status_code, 500)
        gaException = self.getGa4ghException(response.data)
        self.assertEquals(gaException.message, exceptions.ServerError.message)
        self.assertTrue(frontend.app.log_exception.called)

    def testNotImplementedException(self):
        message = "A string unlikely to occur at random."
        exception = exceptions.NotImplementedException(message)
        response = frontend.handleException(exception)
        self.assertEquals(response.status_code, 501)
        gaException = self.getGa4ghException(response.data)
        self.assertEquals(gaException.message, message)
        self.assertFalse(frontend.app.log_exception.called)


def isClassAndBaseServerExceptionSubclass(class_):
    return (inspect.isclass(class_) and
            issubclass(class_, exceptions.BaseServerException))


class TestExceptionConsistency(unittest.TestCase):
    """
    Ensure invariants of exceptions:
    - every exception has a non-None error code
    - every exception has a unique error code
    - every exception has an error code in the range [0, 2**31-1]
    - every exception can be instantiated successfully
    """
    def _getExceptionClasses(self):
        classes = inspect.getmembers(
            exceptions, isClassAndBaseServerExceptionSubclass)
        return [class_ for _, class_ in classes]

    def testCodeInvariants(self):
        codes = set()
        for class_ in self._getExceptionClasses():
            code = class_.getErrorCode()
            self.assertNotIn(code, codes)
            codes.add(code)
            self.assertIsNotNone(code)
            self.assertGreaterEqual(code, 0)
            self.assertLessEqual(code, 2**31 - 1)

    def testInstantiation(self):
        for class_ in self._getExceptionClasses():
            # some exceptions are becoming too complicated to instantiate
            # like the rest of the exceptions; just do them manually
            if class_ == exceptions.RequestValidationFailureException:
                objClass = protocol.SearchReadsRequest
                obj = objClass()
                obj.start = -1
                jsonDict = protocol.toJsonDict(obj)
                args = (jsonDict, objClass)
            else:
                numInitArgs = len(inspect.getargspec(
                    class_.__init__).args) - 1
                args = ['arg' for _ in range(numInitArgs)]
            instance = class_(*args)
            self.assertIsInstance(instance, exceptions.BaseServerException)
            message = instance.getMessage()
            self.assertIsInstance(message, basestring)
            self.assertGreater(len(message), 0)
            self.assertEqual(instance.getErrorCode(), class_.getErrorCode())

    def testGetExceptionClass(self):
        for class_ in self._getExceptionClasses():
            code = class_.getErrorCode()
            self.assertEqual(class_, exceptions.getExceptionClass(code))


@unittest.skip("Protobuf already does validation")
class TestValidationExceptions(unittest.TestCase):
    """
    Tests for exceptions that occur when validation fails.
    """

    def testValidationFailureExceptionMessages(self):
        # RequestValidationFailureException
        wrongString = "thisIsWrong"
        objClass = protocol.SearchReadsRequest
        obj = objClass()
        obj.start = wrongString
        jsonDict = obj.toJsonDict()
        instance = exceptions.RequestValidationFailureException(
            jsonDict, objClass)
        self.assertIn("invalid fields:", instance.message)
        self.assertIn("u'start': u'thisIsWrong'", instance.message)
        self.assertEqual(instance.message.count(wrongString), 2)



"""
The Flask frontend for the GA4GH API.

TODO Document properly.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import datetime
import socket
import urlparse
import functools

import flask
import flask.ext.cors as cors
import humanize
import werkzeug
import oic
import oic.oauth2
import oic.oic.message as message
import requests
import logging
from logging import StreamHandler
from werkzeug.contrib.cache import FileSystemCache

import ga4gh.server
import ga4gh.server.backend as backend
import ga4gh.server.datamodel as datamodel
import ga4gh.server.exceptions as exceptions
import ga4gh.server.datarepo as datarepo
import ga4gh.server.auth as auth
import ga4gh.server.network as network

import ga4gh.schemas.protocol as protocol

MIMETYPE = "application/json"
SEARCH_ENDPOINT_METHODS = ['POST', 'OPTIONS']
SECRET_KEY_LENGTH = 24

app = flask.Flask(__name__)
assert not hasattr(app, 'urls')
app.urls = []

requires_auth = auth.auth_decorator(app)


class NoConverter(werkzeug.routing.BaseConverter):
    """
    A converter that allows the routing matching algorithm to not
    match on certain literal terms

    This is needed because if there are e.g. two routes:

    /callsets/search
    /callsets/<id>

    A request for /callsets/search will get routed to
    the second, which is not what we want.
    """
    def __init__(self, map, *items):
        werkzeug.routing.BaseConverter.__init__(self, map)
        self.items = items

    def to_python(self, value):
        if value in self.items:
            raise werkzeug.routing.ValidationError()
        return value


app.url_map.converters['no'] = NoConverter


class ServerStatus(object):
    """
    Generates information about the status of the server for display
    """
    def __init__(self):
        self.startupTime = datetime.datetime.now()

    def getConfiguration(self):
        """
        Returns a list of configuration (key, value) tuples
        that are useful for users to view on an information page.
        Note that we should be careful here not to leak sensitive
        information. For example, keys and paths of data files should
        not be returned.
        """
        # TODO what other config keys are appropriate to export here?
        keys = [
            'DEBUG', 'REQUEST_VALIDATION',
            'DEFAULT_PAGE_SIZE', 'MAX_RESPONSE_LENGTH', 'LANDING_MESSAGE_HTML'
        ]
        return [(k, app.config[k]) for k in keys]

    def getPreciseUptime(self):
        """
        Returns the server precisely.
        """
        return self.startupTime.strftime("%H:%M:%S %d %b %Y")

    def getLandingMessageHtml(self):
        filePath = app.config.get('LANDING_MESSAGE_HTML')
        try:
            htmlFile = open(filePath, 'r')
            html = htmlFile.read()
            htmlFile.close()
        except:
            html = flask.render_template("landing_message.html")
        return html

    def getNaturalUptime(self):
        """
        Returns the uptime in a human-readable format.
        """
        return humanize.naturaltime(self.startupTime)

    def getProtocolVersion(self):
        """
        Returns the GA4GH protocol version we support.
        """
        return protocol.version

    def getServerVersion(self):
        """
        Returns the software version of this server.
        """
        return ga4gh.server.__version__

    def getUrls(self):
        """
        Returns the list of (httpMethod, URL) tuples that this server
        supports.
        """
        app.urls.sort()
        return app.urls

    def getDatasets(self):
        """
        Returns the list of datasetIds for this backend
        """
        return app.backend.getDataRepository().getDatasets()

    def getVariantSets(self, datasetId):
        """
        Returns the list of variant sets for the dataset
        """
        return app.backend.getDataRepository().getDataset(
            datasetId).getVariantSets()

    def getFeatureSets(self, datasetId):
        """
        Returns the list of feature sets for the dataset
        """
        return app.backend.getDataRepository().getDataset(
            datasetId).getFeatureSets()

    def getContinuousSets(self, datasetId):
        """
        Returns the list of continuous sets for the dataset
        """
        return app.backend.getDataRepository().getDataset(
            datasetId).getContinuousSets()

    def getReadGroupSets(self, datasetId):
        """
        Returns the list of ReadGroupSets for the dataset
        """
        return app.backend.getDataRepository().getDataset(
            datasetId).getReadGroupSets()

    def getReferenceSets(self):
        """
        Returns the list of ReferenceSets for this server.
        """
        return app.backend.getDataRepository().getReferenceSets()

    def getVariantAnnotationSets(self, datasetId):
        """
        Returns the list of ReferenceSets for this server.
        """
        # TODO this should be displayed per-variant set, not per dataset.
        variantAnnotationSets = []
        dataset = app.backend.getDataRepository().getDataset(datasetId)
        for variantSet in dataset.getVariantSets():
            variantAnnotationSets.extend(
                variantSet.getVariantAnnotationSets())
        return variantAnnotationSets

    def getPhenotypeAssociationSets(self, datasetId):
        return app.backend.getDataRepository().getDataset(
            datasetId).getPhenotypeAssociationSets()

    def getRnaQuantificationSets(self, datasetId):
        """
        Returns the list of RnaQuantificationSets for this server.
        """
        return app.backend.getDataRepository().getDataset(
            datasetId).getRnaQuantificationSets()


def reset():
    """
    Resets the flask app; used in testing
    """
    app.config.clear()
    configStr = 'ga4gh.server.serverconfig:FlaskDefaultConfig'
    app.config.from_object(configStr)


def _configure_backend(app):
    """A helper function used just to help modularize the code a bit."""
    # Allocate the backend
    # We use URLs to specify the backend. Currently we have file:// URLs (or
    # URLs with no scheme) for the SqlDataRepository, and special empty:// and
    # simulated:// URLs for empty or simulated data sources.
    dataSource = urlparse.urlparse(app.config["DATA_SOURCE"], "file")

    if dataSource.scheme == "simulated":
        # Ignore the query string
        randomSeed = app.config["SIMULATED_BACKEND_RANDOM_SEED"]
        numCalls = app.config["SIMULATED_BACKEND_NUM_CALLS"]
        variantDensity = app.config["SIMULATED_BACKEND_VARIANT_DENSITY"]
        numVariantSets = app.config["SIMULATED_BACKEND_NUM_VARIANT_SETS"]
        numReferenceSets = app.config[
            "SIMULATED_BACKEND_NUM_REFERENCE_SETS"]
        numReferencesPerReferenceSet = app.config[
            "SIMULATED_BACKEND_NUM_REFERENCES_PER_REFERENCE_SET"]
        numAlignmentsPerReadGroup = app.config[
            "SIMULATED_BACKEND_NUM_ALIGNMENTS_PER_READ_GROUP"]
        numReadGroupsPerReadGroupSet = app.config[
            "SIMULATED_BACKEND_NUM_READ_GROUPS_PER_READ_GROUP_SET"]
        numPhenotypeAssociations = app.config[
            "SIMULATED_BACKEND_NUM_PHENOTYPE_ASSOCIATIONS"]
        numPhenotypeAssociationSets = app.config[
            "SIMULATED_BACKEND_NUM_PHENOTYPE_ASSOCIATION_SETS"]
        numRnaQuantSets = app.config[
            "SIMULATED_BACKEND_NUM_RNA_QUANTIFICATION_SETS"]
        numExpressionLevels = app.config[
            "SIMULATED_BACKEND_NUM_EXPRESSION_LEVELS_PER_RNA_QUANT_SET"]

        dataRepository = datarepo.SimulatedDataRepository(
            randomSeed=randomSeed, numCalls=numCalls,
            variantDensity=variantDensity, numVariantSets=numVariantSets,
            numReferenceSets=numReferenceSets,
            numReferencesPerReferenceSet=numReferencesPerReferenceSet,
            numReadGroupsPerReadGroupSet=numReadGroupsPerReadGroupSet,
            numAlignments=numAlignmentsPerReadGroup,
            numPhenotypeAssociations=numPhenotypeAssociations,
            numPhenotypeAssociationSets=numPhenotypeAssociationSets,
            numRnaQuantSets=numRnaQuantSets,
            numExpressionLevels=numExpressionLevels)
    elif dataSource.scheme == "empty":
        dataRepository = datarepo.EmptyDataRepository()
    elif dataSource.scheme == "file":
        path = os.path.join(dataSource.netloc, dataSource.path)
        dataRepository = datarepo.SqlDataRepository(path)
        dataRepository.open(datarepo.MODE_READ)
    else:
        raise exceptions.ConfigurationException(
            "Unsupported data source scheme: " + dataSource.scheme)
    theBackend = backend.Backend(dataRepository)
    theBackend.setRequestValidation(app.config["REQUEST_VALIDATION"])
    theBackend.setDefaultPageSize(app.config["DEFAULT_PAGE_SIZE"])
    theBackend.setMaxResponseLength(app.config["MAX_RESPONSE_LENGTH"])
    return theBackend


def configure(configFile=None, baseConfig="ProductionConfig",
              port=8000, extraConfig={}):
    """
    TODO Document this critical function! What does it do? What does
    it assume?
    """
    file_handler = StreamHandler()
    file_handler.setLevel(logging.WARNING)
    app.logger.addHandler(file_handler)
    configStr = 'ga4gh.server.serverconfig:{0}'.format(baseConfig)
    app.config.from_object(configStr)
    if os.environ.get('GA4GH_CONFIGURATION') is not None:
        app.config.from_envvar('GA4GH_CONFIGURATION')
    if configFile is not None:
        app.config.from_pyfile(configFile)
    app.config.update(extraConfig.items())
    # Setup file handle cache max size
    datamodel.fileHandleCache.setMaxCacheSize(
        app.config["FILE_HANDLE_CACHE_MAX_SIZE"])
    # Setup CORS
    try:
        cors.CORS(app, allow_headers='Content-Type')
    except AssertionError:
        pass
    app.serverStatus = ServerStatus()

    app.backend = _configure_backend(app)
    if app.config.get('SECRET_KEY'):
        app.secret_key = app.config['SECRET_KEY']
    elif app.config.get('OIDC_PROVIDER'):
        raise exceptions.ConfigurationException(
            'OIDC configuration requires a secret key')
    if app.config.get('CACHE_DIRECTORY'):
        app.cache_dir = app.config['CACHE_DIRECTORY']
    else:
        app.cache_dir = '/tmp/ga4gh'
    app.cache = FileSystemCache(
        app.cache_dir, threshold=5000, default_timeout=600, mode=384)
    # Peer service initialization
    network.initialize(
        app.config.get('INITIAL_PEERS'),
        app.backend.getDataRepository(),
        app.logger)
    app.oidcClient = None
    app.myPort = port
    if app.config.get('AUTH0_ENABLED'):
        emails = app.config.get('AUTH0_AUTHORIZED_EMAILS', '').split(',')
        [auth.authorize_email(e, app.cache) for e in emails]
    if "OIDC_PROVIDER" in app.config:
        # The oic client. If we're testing, we don't want to verify
        # SSL certificates
        app.oidcClient = oic.oic.Client(
            verify_ssl=('TESTING' not in app.config))
        try:
            app.oidcClient.provider_config(app.config['OIDC_PROVIDER'])
        except requests.exceptions.ConnectionError:
            configResponse = message.ProviderConfigurationResponse(
                issuer=app.config['OIDC_PROVIDER'],
                authorization_endpoint=app.config['OIDC_AUTHZ_ENDPOINT'],
                token_endpoint=app.config['OIDC_TOKEN_ENDPOINT'],
                revocation_endpoint=app.config['OIDC_TOKEN_REV_ENDPOINT'])
            app.oidcClient.handle_provider_config(configResponse,
                                                  app.config['OIDC_PROVIDER'])

        # The redirect URI comes from the configuration.
        # If we are testing, then we allow the automatic creation of a
        # redirect uri if none is configured
        redirectUri = app.config.get('OIDC_REDIRECT_URI')
        if redirectUri is None and app.config.get('TESTING'):
            redirectUri = 'https://{0}:{1}/oauth2callback'.format(
                socket.gethostname(), app.myPort)
        app.oidcClient.redirect_uris = [redirectUri]
        if redirectUri is []:
            raise exceptions.ConfigurationException(
                'OIDC configuration requires a redirect uri')

        # We only support dynamic registration while testing.
        if ('registration_endpoint' in app.oidcClient.provider_info and
           app.config.get('TESTING')):
            app.oidcClient.register(
                app.oidcClient.provider_info["registration_endpoint"],
                redirect_uris=[redirectUri])
        else:
            response = message.RegistrationResponse(
                client_id=app.config['OIDC_CLIENT_ID'],
                client_secret=app.config['OIDC_CLIENT_SECRET'],
                redirect_uris=[redirectUri],
                verify_ssl=False)
            app.oidcClient.store_registration_info(response)


def getFlaskResponse(responseString, httpStatus=200):
    """
    Returns a Flask response object for the specified data and HTTP status.
    """
    return flask.Response(responseString, status=httpStatus, mimetype=MIMETYPE)


def handleHttpPost(request, endpoint):
    """
    Handles the specified HTTP POST request, which maps to the specified
    protocol handler endpoint and protocol request class.
    """
    if request.mimetype and request.mimetype != MIMETYPE:
        raise exceptions.UnsupportedMediaTypeException()
    request = request.get_data()
    if request == '' or request is None:
        request = '{}'
    responseStr = endpoint(request)
    return getFlaskResponse(responseStr)


def handleList(endpoint, request):
    """
    Handles the specified HTTP GET request, mapping to a list request
    """
    responseStr = endpoint(request.get_data())
    return getFlaskResponse(responseStr)


def handleHttpGet(id_, endpoint):
    """
    Handles the specified HTTP GET request, which maps to the specified
    protocol handler endpoint and protocol request class
    """
    responseStr = endpoint(id_)
    return getFlaskResponse(responseStr)


def handleHttpOptions():
    """
    Handles the specified HTTP OPTIONS request.
    """
    response = getFlaskResponse("")
    response.headers.add("Access-Control-Request-Methods", "GET,POST,OPTIONS")
    return response


@app.errorhandler(Exception)
def handleException(exception):
    """
    Handles an exception that occurs somewhere in the process of handling
    a request.
    """
    serverException = exception
    if not isinstance(exception, exceptions.BaseServerException):
        with app.test_request_context():
            app.log_exception(exception)
        serverException = exceptions.getServerError(exception)
    error = serverException.toProtocolElement()
    # If the exception is being viewed by a web browser, we can render a nicer
    # view.
    if flask.request and 'Accept' in flask.request.headers and \
            flask.request.headers['Accept'].find('text/html') != -1:
        message = "<h1>Error {}</h1><pre>{}</pre>".format(
                    serverException.httpStatus,
                    protocol.toJson(error))
        if serverException.httpStatus == 401 \
                or serverException.httpStatus == 403:
            message += "Please try <a href=\"/login\">logging in</a>."
        return message
    else:
        responseStr = protocol.toJson(error)
        return getFlaskResponse(responseStr, serverException.httpStatus)


def startLogin():
    """
    If we are not logged in, this generates the redirect URL to the OIDC
    provider and returns the redirect response
    :return: A redirect response to the OIDC provider
    """
    flask.session["state"] = oic.oauth2.rndstr(SECRET_KEY_LENGTH)
    flask.session["nonce"] = oic.oauth2.rndstr(SECRET_KEY_LENGTH)
    args = {
        "client_id": app.oidcClient.client_id,
        "response_type": "code",
        "scope": ["openid", "profile"],
        "nonce": flask.session["nonce"],
        "redirect_uri": app.oidcClient.redirect_uris[0],
        "state": flask.session["state"]
    }

    result = app.oidcClient.do_authorization_request(
        request_args=args, state=flask.session["state"])
    return flask.redirect(result.url)


@app.before_request
def checkAuthentication():
    """
    The request will have a parameter 'key' if it came from the command line
    client, or have a session key of 'key' if it's the browser.
    If the token is not found, start the login process.

    If there is no oidcClient, we are running naked and we don't check.
    If we're being redirected to the oidcCallback we don't check.

    :returns None if all is ok (and the request handler continues as usual).
    Otherwise if the key was in the session (therefore we're in a browser)
    then startLogin() will redirect to the OIDC provider. If the key was in
    the request arguments, we're using the command line and just raise an
    exception.
    """
    if app.oidcClient is None:
        return
    if flask.request.endpoint == 'oidcCallback':
        return
    key = flask.session.get('key') or flask.request.args.get('key')
    if key is None or not app.cache.get(key):
        if 'key' in flask.request.args:
            raise exceptions.NotAuthenticatedException()
        else:
            return startLogin()


def handleFlaskGetRequest(id_, flaskRequest, endpoint):
    """
    Handles the specified flask request for one of the GET URLs
    Invokes the specified endpoint to generate a response.
    """
    if flaskRequest.method == "GET":
        return handleHttpGet(id_, endpoint)
    else:
        raise exceptions.MethodNotAllowedException()


def handleFlaskListRequest(id_, flaskRequest, endpoint):
    """
    Handles the specified flask list request for one of the GET URLs.
    Invokes the specified endpoint to generate a response.
    """

    return handleList(endpoint, flaskRequest)


def handleFlaskPostRequest(flaskRequest, endpoint):
    """
    Handles the specified flask request for one of the POST URLS
    Invokes the specified endpoint to generate a response.
    """
    if flaskRequest.method == "POST":
        return handleHttpPost(flaskRequest, endpoint)
    elif flaskRequest.method == "OPTIONS":
        return handleHttpOptions()
    else:
        raise exceptions.MethodNotAllowedException()


class DisplayedRoute(object):
    """
    Registers that a route should be displayed on the html page
    """
    def __init__(
            self, path, postMethod=False, pathDisplay=None):
        self.path = path
        self.methods = None
        if postMethod:
            methodDisplay = 'POST'
            self.methods = SEARCH_ENDPOINT_METHODS
        else:
            methodDisplay = 'GET'
        if pathDisplay is None:
            pathDisplay = path
        app.urls.append((methodDisplay, pathDisplay))

    def __call__(self, func):
        if self.methods is None:
            app.add_url_rule(self.path, func.func_name, func)
        else:
            app.add_url_rule(
                self.path, func.func_name, func, methods=self.methods)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            return result
        return wrapper


@app.route('/')
def index():
    response = flask.render_template('index.html',
                                     info=app.serverStatus)
    if app.config.get('AUTH0_ENABLED'):
        key = (flask.request.args.get('key'))
        try:
            print(key)
            profile = app.cache.get(key)
        except:
            raise exceptions.NotAuthorizedException()
        if (profile):
            return response
        else:
            exceptions.NotAuthenticatedException()
    else:
        return response


@app.route("/login")
def login():
    if app.config.get('AUTH0_ENABLED'):
        if (flask.request.args.get('code')):
            return auth.render_key(app, key=flask.request.args.get('code'))
        else:
            return auth.render_login(
                app=app,
                scopes=app.config.get('AUTH0_SCOPES'),
                redirect_uri=app.config.get('AUTH0_CALLBACK_URL'),
                domain=app.config.get('AUTH0_HOST'),
                client_id=app.config.get('AUTH0_CLIENT_ID'))
    else:
        raise exceptions.NotFoundException()


@app.route('/callback')
def callback_handling():
    if app.config.get('AUTH0_ENABLED'):
        return auth.callback_maker(
            cache=app.cache,
            domain=app.config.get('AUTH0_HOST'),
            client_id=app.config.get('AUTH0_CLIENT_ID'),
            client_secret=app.config.get('AUTH0_CLIENT_SECRET'),
            redirect_uri=app.config.get('AUTH0_CALLBACK_URL'))()
    else:
        raise exceptions.NotFoundException()


@app.route("/logout")
@requires_auth
@cors.cross_origin(headers=['Content-Type', 'Authorization'])
def logout():
    key = flask.session['auth0_key']
    auth.logout(app.cache)
    url = 'https://{}/v2/logout?access_token={}&?client_id={}'.format(
        app.config.get('AUTH0_HOST'),
        key,
        app.config.get('AUTH0_CLIENT_ID'),
        app.config.get('AUTH0_CALLBACK_URL'))
    return flask.redirect(url)


@app.route('/favicon.ico')
@app.route('/robots.txt')
def robots():
    return flask.send_from_directory(
        app.static_folder, flask.request.path[1:])


@DisplayedRoute('/info')
@requires_auth
def getInfo():
    return handleFlaskGetRequest(
        None, flask.request, app.backend.runGetInfo)


@DisplayedRoute('/references/<id>')
@requires_auth
def getReference(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetReference)


@DisplayedRoute('/referencesets/<id>')
def getReferenceSet(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetReferenceSet)


@DisplayedRoute('/listreferencebases', postMethod=True)
def listReferenceBases():
    return handleFlaskListRequest(
        id, flask.request, app.backend.runListReferenceBases)


@DisplayedRoute('/callsets/search', postMethod=True)
def searchCallSets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchCallSets)


@DisplayedRoute('/readgroupsets/search', postMethod=True)
def searchReadGroupSets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchReadGroupSets)


@DisplayedRoute('/reads/search', postMethod=True)
def searchReads():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchReads)


@DisplayedRoute('/referencesets/search', postMethod=True)
def searchReferenceSets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchReferenceSets)


@DisplayedRoute('/references/search', postMethod=True)
def searchReferences():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchReferences)


@DisplayedRoute('/variantsets/search', postMethod=True)
def searchVariantSets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchVariantSets)


@DisplayedRoute('/variants/search', postMethod=True)
def searchVariants():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchVariants)


@DisplayedRoute('/variantannotationsets/search', postMethod=True)
def searchVariantAnnotationSets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchVariantAnnotationSets)


@DisplayedRoute('/variantannotations/search', postMethod=True)
def searchVariantAnnotations():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchVariantAnnotations)


@DisplayedRoute('/datasets/search', postMethod=True)
@requires_auth
def searchDatasets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchDatasets)


@DisplayedRoute('/featuresets/search', postMethod=True)
@requires_auth
def searchFeatureSets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchFeatureSets)


@DisplayedRoute('/features/search', postMethod=True)
@requires_auth
def searchFeatures():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchFeatures)


@DisplayedRoute('/continuoussets/search', postMethod=True)
@requires_auth
def searchContinuousSets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchContinuousSets)


@DisplayedRoute('/continuous/search', postMethod=True)
@requires_auth
def searchContinuous():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchContinuous)


@DisplayedRoute('/biosamples/search', postMethod=True)
@requires_auth
def searchBiosamples():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchBiosamples)


@DisplayedRoute('/individuals/search', postMethod=True)
@requires_auth
def searchIndividuals():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchIndividuals)


@DisplayedRoute('/peers/list', postMethod=True)
@requires_auth
def listPeers():
    return handleFlaskPostRequest(
        flask.request, app.backend.runListPeers)


@DisplayedRoute('/announce', postMethod=True)
@requires_auth
def announce():
    # We can't use the post handler here because we want detailed request
    # data.
    return app.backend.runAddAnnouncement(flask.request)


@DisplayedRoute(
    '/biosamples/<no(search):id>',
    pathDisplay='/biosamples/<id>')
@requires_auth
def getBiosample(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetBiosample)


@DisplayedRoute(
    '/individuals/<no(search):id>',
    pathDisplay='/individuals/<id>')
@requires_auth
def getIndividual(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetIndividual)


@DisplayedRoute('/rnaquantificationsets/search', postMethod=True)
@requires_auth
def searchRnaQuantificationSets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchRnaQuantificationSets)


@DisplayedRoute('/rnaquantifications/search', postMethod=True)
@requires_auth
def searchRnaQuantifications():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchRnaQuantifications)


@DisplayedRoute('/expressionlevels/search', postMethod=True)
@requires_auth
def searchExpressionLevels():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchExpressionLevels)


@DisplayedRoute(
    '/variantsets/<no(search):id>',
    pathDisplay='/variantsets/<id>')
@requires_auth
def getVariantSet(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetVariantSet)


@DisplayedRoute(
    '/variants/<no(search):id>',
    pathDisplay='/variants/<id>')
@requires_auth
def getVariant(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetVariant)


@DisplayedRoute(
    '/readgroupsets/<no(search):id>',
    pathDisplay='/readgroupsets/<id>')
@requires_auth
def getReadGroupSet(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetReadGroupSet)


@DisplayedRoute('/readgroups/<id>')
@requires_auth
def getReadGroup(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetReadGroup)


@DisplayedRoute(
    '/callsets/<no(search):id>',
    pathDisplay='/callsets/<id>')
@requires_auth
def getCallSet(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetCallSet)


@DisplayedRoute(
    '/featuresets/<no(search):id>',
    pathDisplay='/featuresets/<id>')
@requires_auth
def getFeatureSet(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetFeatureSet)


@DisplayedRoute(
    '/features/<no(search):id>',
    pathDisplay='/features/<id>')
@requires_auth
def getFeature(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetFeature)


@DisplayedRoute(
    '/continuoussets/<no(search):id>',
    pathDisplay='/continuoussets/<id>')
@requires_auth
def getcontinuousSet(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetContinuousSet)


@DisplayedRoute(
    '/rnaquantificationsets/<no(search):id>',
    pathDisplay='/rnaquantificationsets/<id>')
@requires_auth
def getRnaQuantificationSet(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetRnaQuantificationSet)


@DisplayedRoute(
    '/rnaquantifications/<no(search):id>',
    pathDisplay='/rnaquantifications/<id>')
@requires_auth
def getRnaQuantification(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetRnaQuantification)


@DisplayedRoute(
    '/expressionlevels/<no(search):id>',
    pathDisplay='/expressionlevels/<id>')
@requires_auth
def getExpressionLevel(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetExpressionLevel)


@app.route('/oauth2callback', methods=['GET'])
def oidcCallback():
    """
    Once the authorization provider has cleared the user, the browser
    is returned here with a code. This function takes that code and
    checks it with the authorization provider to prove that it is valid,
    and get a bit more information about the user (which we don't use).

    A token is generated and given to the user, and the authorization info
    retrieved above is stored against this token. Later, when a client
    connects with this token, it is assumed to be a valid user.

    :return: A display of the authentication token to use in the client. If
    OIDC is not configured, raises a NotImplementedException.
    """
    if app.oidcClient is None:
        raise exceptions.NotImplementedException()
    response = dict(flask.request.args.iteritems(multi=True))
    aresp = app.oidcClient.parse_response(
        message.AuthorizationResponse,
        info=response,
        sformat='dict')
    sessState = flask.session.get('state')
    respState = aresp['state']
    if (not isinstance(aresp, message.AuthorizationResponse) or
            respState != sessState):
        raise exceptions.NotAuthenticatedException()

    args = {
        "code": aresp['code'],
        "redirect_uri": app.oidcClient.redirect_uris[0],
        "client_id": app.oidcClient.client_id,
        "client_secret": app.oidcClient.client_secret
    }
    atr = app.oidcClient.do_access_token_request(
        scope="openid",
        state=respState,
        request_args=args)

    if not isinstance(atr, message.AccessTokenResponse):
        raise exceptions.NotAuthenticatedException()

    atrDict = atr.to_dict()
    if flask.session.get('nonce') != atrDict['id_token']['nonce']:
        raise exceptions.NotAuthenticatedException()
    key = oic.oauth2.rndstr(SECRET_KEY_LENGTH)
    flask.session['key'] = key
    token_data = aresp["code"], respState, atrDict
    app.cache.set(key, token_data)
    # flask.url_for is broken. It relies on SERVER_NAME for both name
    # and port, and defaults to 'localhost' if not found. Therefore
    # we need to fix the returned url
    indexUrl = flask.url_for('index', _external=True)
    indexParts = list(urlparse.urlparse(indexUrl))
    if ':' not in indexParts[1]:
        indexParts[1] = '{}:{}'.format(socket.gethostname(), app.myPort)
        indexUrl = urlparse.urlunparse(indexParts)
    response = flask.redirect(indexUrl)
    return response


@DisplayedRoute(
    '/datasets/<no(search):id>',
    pathDisplay='/datasets/<id>')
@requires_auth
def getDataset(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetDataset)


@DisplayedRoute(
    '/variantannotationsets/<no(search):id>',
    pathDisplay='/variantannotationsets/<id>')
@requires_auth
def getVariantAnnotationSet(id):
    return handleFlaskGetRequest(
        id, flask.request, app.backend.runGetVariantAnnotationSet)


@DisplayedRoute('/phenotypes/search', postMethod=True)
@requires_auth
def searchPhenotypes():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchPhenotypes)


@DisplayedRoute('/featurephenotypeassociations/search', postMethod=True)
@requires_auth
def searchGenotypePhenotypes():
    return handleFlaskPostRequest(
        flask.request,
        app.backend.runSearchGenotypePhenotypes)


@DisplayedRoute('/phenotypeassociationsets/search', postMethod=True)
@requires_auth
def searchPhenotypeAssociationSets():
    return handleFlaskPostRequest(
        flask.request, app.backend.runSearchPhenotypeAssociationSets)


# The below methods ensure that JSON is returned for various errors
# instead of the default, html


@app.errorhandler(401)
def unauthorizedHandler(errorString):
    return handleException(exceptions.UnauthorizedException(errorString))


@app.errorhandler(404)
def pathNotFoundHandler(errorString):
    return handleException(exceptions.PathNotFoundException())


@app.errorhandler(405)
def methodNotAllowedHandler(errorString):
    return handleException(exceptions.MethodNotAllowedException())


@app.errorhandler(403)
def notAuthenticatedHandler(errorString):
    return handleException(exceptions.NotAuthenticatedException(errorString))



"""
Exceptions for the GA4GH server. Each exception that can occur in the server
is given a unique error code that is derived from its name.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import zlib
import inspect
import json

import ga4gh.schemas.protocol as protocol


def getExceptionClass(errorCode):
    """
    Converts the specified error code into the corresponding class object.
    Raises a KeyError if the errorCode is not found.
    """
    classMap = {}
    for name, class_ in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(class_) and issubclass(class_, BaseServerException):
            classMap[class_.getErrorCode()] = class_
    return classMap[errorCode]


def getServerError(exception):
    """
    Converts the specified exception that is not a subclass of
    BaseServerException into a ServerError so that it can be correctly
    serialised and communicated back to the client.
    """
    assert not isinstance(exception, BaseServerException)
    serverException = ServerError()
    return serverException


class BaseServerException(Exception):
    """
    Superclass of all exceptions that can occur in the GA4GH reference
    server.
    """
    message = "Error code not set in exception; this is a bug."

    def __init__(self, *args, **kwargs):
        # We have this constructor so that we can always create an
        # instance of our base classes using inspect. This is useful
        # for testing.
        super(BaseServerException, self).__init__(*args, **kwargs)

    def getMessage(self):
        """
        Returns the message that we map into the GA4GH protocol message.
        For many exceptions we can simply set the message statically
        for a given class, and this message will be returned by
        default. For more elaborate messages that must be contructed
        at run time, we can override this method and use state held
        in instance variables in the exception.
        """
        return self.message

    @classmethod
    def getErrorCode(cls):
        """
        Returns an integer that can be used to identify this class.
        """
        # We use the CRC32 digest of the class name as a unique code.
        # We follow the recommendation of the Python docs to ensure
        # that this value is signed 32 bit integer, and then mod it
        # to ensure non-negativity
        code = (zlib.crc32(cls.__name__) & 0xffffffff) % 2**31
        return code

    def __str__(self):
        return self.getMessage()


#####################################################################
#
# Exceptions that occur in the normal operation of the server
#
#####################################################################


class RuntimeException(BaseServerException):
    """
    exceptions that can occur during the processing of a client request
    """
    httpStatus = -1
    message = "Error code not set in exception; this is a bug."

    def toProtocolElement(self):
        """
        Converts this exception into the GA4GH protocol type so that
        it can be communicated back to the client.
        """
        error = protocol.GAException()
        error.error_code = self.getErrorCode()
        error.message = self.getMessage()
        return error


class BadRequestException(RuntimeException):
    """
    A request that we don't like was sent to the server.
    """
    httpStatus = 400
    message = "Bad request"


class BadRequestIntegerException(BadRequestException):
    def __init__(self, attrName, intString):
        self.message = \
            "{} argument '{}' could not be parsed as an integer".format(
                attrName, intString)


class BadPageSizeException(BadRequestException):
    def __init__(self, pageSize):
        self.message = "Request page size '{}' is invalid".format(pageSize)


class BadPageTokenException(BadRequestException):
    message = "Request page token invalid"


class BadIdentifierException(BadRequestException):
    def __init__(self, id_, msg=None):
        self.message = "The identifier provided is invalid: '{}' ".format(id_)
        if msg is not None:
            self.message += msg


class BadIdentifierNotStringException(BadIdentifierException):
    def __init__(self, localId):
        self.message = (
            "The identifier provided is not a string: '{}'".format(
                localId))


class InvalidJsonException(BadRequestException):
    def __init__(self, jsonString):
        self.message = "Cannot parse JSON: '{}'".format(jsonString)


class Validator(object):
    """
    Check that a JSON dictionary is a valid representation of a protocol
    buffer class
    """

    def __init__(self, class_):
        self.class_ = class_

    def getInvalidFields(self, jsonDict):
        # FIXME: get the proper list of fields
        protocol.fromJson(json.dumps(jsonDict), self.class_)
        return []


class RequestValidationFailureException(BadRequestException):
    """
    A validation of the request data failed
    """
    def __init__(self, jsonDict, requestClass):
        messageString = (
            "Request '{}' is not a valid instance of {}; "
            "invalid fields: {}")
        validator = Validator(requestClass)
        self.message = messageString.format(
            jsonDict, requestClass,
            validator.getInvalidFields(jsonDict)
        )


class BadFeatureSetSearchRequestRegularExpression(BadRequestException):
    message = "Malformed regular expression"
    httpStatus = 400


class DatamodelValidationException(BadRequestException):
    """
    Some bad data was passed to us by the client that made no sense
    in the context of a particular datamodel object
    """


class BadUrlException(RuntimeException):
    def __init__(self, url):
        self.message = "The URL: '{}' was malformed".format(url)
        self.httpStatus = 400


class ReadGroupSetNotMappedToReferenceSetException(BadRequestException):

    def __init__(self, readGroupSetId):
        self.message = (
            "ReadGroupSet '{}' is not mapped to any referenceSet".format(
                readGroupSetId))


class NotFoundException(RuntimeException):
    """
    The superclass of all exceptions in which some resource was not
    found.
    """
    httpStatus = 404
    message = "A resource was not found"


class PeerNotFoundException(NotFoundException):
    def __init__(self, url):
        self.message = "A peer with url '{}' was not found".format(
            url)


class PathNotFoundException(NotFoundException):
    message = "The request path was not found"


class ConfigurationException(BaseServerException):
    pass


class ObjectNotFoundException(NotFoundException):
    message = "The requested object was not found"


class VariantSetNotFoundException(NotFoundException):
    def __init__(self, variantSetId):
        self.message = "The requested VariantSet '{}' was not found".format(
            variantSetId)


class PhenotypeAssociationSetNotFoundException(NotFoundException):
    def __init__(self, paSetId):
        self.message = (
            "The requested PhenotypeAssociationSet '{}' "
            "was not found".format(
                paSetId))


class BiosampleNotFoundException(NotFoundException):
    def __init__(self, biosampleId):
        self.message = "The requested Biosample '{}' was not found".format(
            biosampleId)


class IndividualNotFoundException(NotFoundException):
    def __init__(self, individualId):
        self.message = "The requested Individual '{}' was not found".format(
            individualId)


class AnnotationSetNotFoundException(NotFoundException):
    def __init__(self, variantAnnotationSetId):
        self.message = "The requested VariantAnnotationSet '{}'" \
            "was not found".format(variantAnnotationSetId)


class CallSetNotFoundException(NotFoundException):
    def __init__(self, callSetId):
        self.message = "The requested CallSet '{}' was not found".format(
            callSetId)


class DatasetNotFoundException(NotFoundException):
    def __init__(self, datasetId):
        self.message = "The requested dataset '{}' was not found".format(
            datasetId)


class ReadGroupNotFoundException(ObjectNotFoundException):
    def __init__(self, readGroupId):
        self.message = "readGroupId '{}' not found".format(readGroupId)


class ReferenceSetNotFoundException(ObjectNotFoundException):
    def __init__(self, referenceSetId):
        self.message = "referenceSetId '{}' not found".format(referenceSetId)


class ReferenceNotFoundException(ObjectNotFoundException):
    def __init__(self, referenceId):
        self.message = "referenceId '{}' not found".format(referenceId)


class OntologyNotFoundException(ObjectNotFoundException):
    def __init__(self, ontologyId):
        self.message = "ontologyId '{}' not found".format(ontologyId)


class ObjectWithIdNotFoundException(ObjectNotFoundException):
    def __init__(self, objectId):
        self.message = "No object of this type exists with id '{}'".format(
            objectId)


class UnsupportedMediaTypeException(RuntimeException):
    httpStatus = 415
    message = "Unsupported media type"


class RangeErrorException(RuntimeException):
    """
    The superclass of all exceptions for which a query range error occured.
    This raises a HTTP Error 416 "Requested Range not satisfiable".
    """
    httpStatus = 416
    message = "Requested Range not satisfiable"


class ReferenceRangeErrorException(RangeErrorException):
    """
    Exception raised when the client attempts to access coordinates
    outside of the reference.
    """
    def __init__(self, referenceId, start, end):
        self.message = (
            "Query ({}, {}) outside of range for reference {}".format(
                start, end, referenceId))


class MethodNotAllowedException(RuntimeException):
    httpStatus = 405
    message = "Method not allowed"


class NotAuthenticatedException(RuntimeException):
    httpStatus = 403

    def __init__(self, message=None):
        if message is None:
            self.message = "Not authenticated. Use the " \
                           "key on the server index page."
        else:
            self.message = message


class NotAuthorizedException(RuntimeException):
    httpStatus = 401

    def __init__(self, message=None):
        if message is None:
            self.message = "Not authenticated. Use the " \
                           "key on the server index page."
        else:
            self.message = message


class NotImplementedException(RuntimeException):
    """
    Exception raised when a part of the API has not been implemented.
    """
    httpStatus = 501

    def __init__(self, message=None):
        if message is None:
            self.message = "Path not implemented"
        else:
            self.message = message


class UnmappedReadsNotSupported(NotImplementedException):
    def __init__(self):
        self.message = (
            "Unmapped reads are not yet supported; "
            "please specify a reference")


class CallSetNotInVariantSetException(NotFoundException):
    """
    Indicates a request was made for a callSet not in the actual variantSet
    """
    def __init__(self, callSetId, variantSetId):
        self.message = "callSet '{0}' not in variantSet '{1}'".format(
            callSetId, variantSetId)


class CallSetNameNotFoundException(NotFoundException):
    """
    Indicates a request was made for a callSet with a name that
    does not exist.
    """
    def __init__(self, name):
        self.message = "CallSet with name '{0}' not found".format(name)


class VariantSetNameNotFoundException(NotFoundException):
    """
    Indicates a request was made for a VariantSet with a name that
    does not exist.
    """
    def __init__(self, name):
        self.message = "VariantSet with name '{0}' not found".format(name)


class ReadGroupSetNameNotFoundException(NotFoundException):
    """
    Indicates a request was made for a ReadGroupSet with a name that
    does not exist.
    """
    def __init__(self, name):
        self.message = "ReadGroupSet with name '{0}' not found".format(name)


class ReferenceNameNotFoundException(NotFoundException):
    """
    Indicates a request was made for a Reference with a name that
    does not exist.
    """
    def __init__(self, name):
        self.message = "Reference with name '{0}' not found".format(name)


class BiosampleNameNotFoundException(NotFoundException):
    """
    Indicates a request was made for a Biosample with a name that
    does not exist.
    """
    def __init__(self, name):
        self.message = "Biosample with name '{0}' not found".format(name)


class IndividualNameNotFoundException(NotFoundException):
    """
    Indicates a request was made for an Individual with a name that
    does not exist.
    """
    def __init__(self, name):
        self.message = "Individual with name '{0}' not found".format(name)


class ReferenceSetNameNotFoundException(NotFoundException):
    """
    Indicates a request was made for a ReferenceSetSet with a name that
    does not exist.
    """
    def __init__(self, name):
        self.message = "ReferenceSet with name '{0}' not found".format(name)


class DatasetNameNotFoundException(NotFoundException):
    """
    Indicates a request was made for a Dataset with a name that
    does not exist.
    """
    def __init__(self, name):
        self.message = "Dataset with name '{0}' not found".format(name)


class OntologyNameNotFoundException(NotFoundException):
    """
    Indicates a request was made for an Ontology with a name that
    does not exist.
    """
    def __init__(self, name):
        self.message = "Ontology with name '{0}' not found".format(name)


class FeatureSetNotFoundException(NotFoundException):
    def __init__(self, featureSetId):
        self.message = (
            "FeatureSet with id '{0}' not found".format(featureSetId))


class FeatureSetNameNotFoundException(NotFoundException):
    def __init__(self, featureSetId):
        self.message = (
            "FeatureSet with name '{0}' not found".format(featureSetId))


class ParentIncompatibleWithFeatureSet(BadRequestException):
    def __init__(self):
        self.message = (
            "Parent feature incompatible with requested Feature Set."
        )


class FeatureSetNotSpecifiedException(BadRequestException):
    def __init__(self):
        self.message = (
            "One of featureSetId or parentId must be supplied."
        )


class ContinuousSetNotFoundException(NotFoundException):
    def __init__(self, continuousSetId):
        self.message = (
            "ContinuousSet with id '{0}' not found".format(continuousSetId))


class ContinuousSetNameNotFoundException(NotFoundException):
    def __init__(self, continuousSetId):
        self.message = (
            "ContinuousSet with name '{0}' not found".format(continuousSetId))


class RnaQuantificationSetNotFoundException(NotFoundException):
    def __init__(self, name):
        self.message = (
            "RnaQuantificationSet with name '{0}' not found".format(name))


class RnaQuantificationSetNameNotFoundException(NotFoundException):
    def __init__(self, name):
        self.message = (
            "RnaQuantificationSet with name '{0}' not found".format(name))


class RnaQuantificationNotFoundException(NotFoundException):
    def __init__(self, name):
        self.message = (
            "RnaQuantification with name '{0}' not found".format(name))


class ExpressionLevelNotFoundException(NotFoundException):
    def __init__(self, name):
        self.message = (
            "ExpressionLevel with name '{0}' not found".format(name))


class DataException(BaseServerException):
    """
    Exceptions thrown during data ingestion.
    """
    message = "xs"


class RepoNotFoundException(DataException):

    def __init__(self, repoPath):
        self.message = (
            "Database file '{}' does not exist.  Either "
            "change the configuration to point to a valid file or "
            "create one using the repo manager.".format(repoPath))


class RepoInvalidDatabaseException(DataException):

    def __init__(self, repoPath):
        self.message = (
            "Database file '{}' is malformed.  Either "
            "change the configuration to point to a valid file or "
            "create one using the repo manager.".format(repoPath))


class RepoSchemaVersionMismatchException(DataException):

    def __init__(self, repoVersion, expectedVersion):
        self.message = (
            "Repository version '{}' does not match expected "
            "version '{}'".format(repoVersion, expectedVersion))


class FileOpenFailedException(DataException):

    def __init__(self, filename):
        self.message = "Failed to open file '{}'".format(filename)


class EmptyDirException(DataException):

    def __init__(self, dirname, filetype):
        self.message = "Directory '{}' empty, no {} file was found".format(
            dirname, filetype)


class OntologyFileFormatException(DataException):
    """
    Exception thrown when an error occurs processing an OBO ontology
    file.
    """
    def __init__(self, filename, message):
        self.message = "Error processing ontology OBO file '{}': {}".format(
            filename, message)


class MalformedException(DataException):
    """
    A base exception class for exceptions thrown when faulty VCF file
    was found, which stops the server to processed variant search.
    """
    message = "Faulty entry found in data file."


class NotIndexedException(MalformedException):
    """
    VCF/BCF files must be indexed, with the same prefix.
    """
    def __init__(self, fileName):
        self.message = (
            "Indexing file is missing, File {} must be"
            " indexed.".format(fileName))


class OverlappingVcfException(MalformedException):
    """
    Exception thrown when two VCF files within a VariantSet directory
    contain records for the same contig.
    """
    def __init__(self, fileName, contig):
        self.message = (
            "VCF file '{}' contains records for contig '{}'. Other files"
            " in this VariantSet have records for this contig, and overlapping"
            " VCFs are not permitted.".format(fileName, contig))


class InconsistentMetaDataException(MalformedException):
    """
    Exception thrown when two VCF files within a VariantSet direcotry
    contain different metadata entries.
    """
    def __init__(self, fileName):
        self.fileName = fileName
        self.message = (
            "Metadata of {} is not consistent with other files"
            " in this VariantSet, and inconsistent metadata is not"
            " permitted".format(self.fileName))


class DuplicateCallSetIdException(MalformedException):
    """
    Exception thrown when the same sample ID occurs multiple times in
    one VCF file
    """
    def __init__(self, fileName, callSetId):
        self.callSetId = callSetId
        self.message = (
            "CallSetId: {} appeared multiple times in file {}. Duplicated"
            " sample ID is not permitted".format(self.callSetId, fileName))


class InconsistentCallSetIdException(MalformedException):
    """
    Exception thrown when two VCF files contain different sample IDs
    within a VariantSet directory.
    """
    def __init__(self, fileName):
        self.message = (
            "Inconsistent sample names found in {}. Sample IDs must be"
            " consistent within the same VariantSet"
            " directory.".format(fileName))


class MultipleReferenceSetsInReadGroupSet(MalformedException):
    """
    A BAM file contains reference sequences from multiple reference
    sets.
    """
    def __init__(self, fileName, referenceSetName, otherReferenceSetName):
        self.message = (
            "The BAM file '{}' contains the referenceSets '{}' and "
            "'{}'; at most one referenceSet per file is allowed.".format(
                fileName, referenceSetName, otherReferenceSetName))


###############################################################
#
# Internal errors. These are exceptions that we regard as bugs.
#
###############################################################


class ServerError(RuntimeException):
    """
    Superclass of all exceptions that indicate a bug has occured.
    """
    httpStatus = 500
    message = "Internal Server Error"


#####################################################################
#
# Repo manager exceptions
#
#####################################################################


class RepoManagerException(Exception):
    """
    Signals something went wrong inside the repo manager
    """


class DuplicateNameException(RepoManagerException):
    """
    The user has tried to insert an object with the same name as
    an existing object within a given container.
    """
    def __init__(self, objectName, containerName=None):
        msg = "Name '{}' already exists".format(objectName)
        if containerName is not None:
            msg += " in '{}'".format(containerName)
        super(DuplicateNameException, self).__init__(msg)


class MissingIndexException(RepoManagerException):
    """
    An index file for a remote BAM/VCF has not been provided.
    """
    def __init__(self, dataUrl):
        msg = "An index file must be provided for remote file '{}'".format(
            dataUrl)
        super(MissingIndexException, self).__init__(msg)


class UnsupportedFormatException(RepoManagerException):
    """
    The user has specified a data format which is not supported.
    """
    def __init__(self, dataFormat):
        msg = "Unsupported format: {}".format(dataFormat)
        super(UnsupportedFormatException, self).__init__(msg)



"""
Tools for alignment files
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import pysam

import ga4gh.common.utils as utils


class AlignmentFileConstants(object):
    """
    A container class for constants dealing with alignment files
    """
    SAM = "SAM"
    BAM = "BAM"
    BAI = "BAI"


class AlignmentFileTool(object):
    """
    Helps with operations on BAM and SAM files
    """
    def __init__(self, inputFileFormat, outputFileFormat):
        self.inputFileFormat = inputFileFormat
        self.outputFileFormat = outputFileFormat
        self.args = None

    def parseArgs(self):
        description = "{} to {} conversion tool".format(
            self.inputFileFormat, self.outputFileFormat)
        parser = argparse.ArgumentParser(
            description=description)
        inputHelpText = "the name of the {} file to read".format(
            self.inputFileFormat)
        parser.add_argument(
            "inputFile", help=inputHelpText)
        outputHelpText = "the name of the {} file to write".format(
            self.outputFileFormat)
        defaultOutputFilePath = "out.{}".format(
            self.outputFileFormat.lower())
        parser.add_argument(
            "--outputFile", "-o", default=defaultOutputFilePath,
            help=outputHelpText)
        parser.add_argument(
            "--numLines", "-n", default=10,
            help="the number of lines to write")
        parser.add_argument(
            "--skipIndexing", default=False, action='store_true',
            help="don't create an index file")
        args = parser.parse_args()
        self.args = args

    def convert(self):
        # set flags
        if self.inputFileFormat == AlignmentFileConstants.SAM:
            inputFlags = "r"
        elif self.inputFileFormat == AlignmentFileConstants.BAM:
            inputFlags = "rb"
        if self.outputFileFormat == AlignmentFileConstants.SAM:
            outputFlags = "wh"
        elif self.outputFileFormat == AlignmentFileConstants.BAM:
            outputFlags = "wb"
        # open files
        inputFile = pysam.AlignmentFile(
            self.args.inputFile, inputFlags)
        outputFile = pysam.AlignmentFile(
            self.args.outputFile, outputFlags, header=inputFile.header)
        outputFilePath = outputFile.filename
        utils.log("Creating alignment file '{}'".format(outputFilePath))
        # write new file
        for _ in xrange(self.args.numLines):
            alignedSegment = inputFile.next()
            outputFile.write(alignedSegment)
        # clean up
        inputFile.close()
        outputFile.close()
        # create index file
        if (not self.args.skipIndexing and
                self.outputFileFormat == AlignmentFileConstants.BAM):
            indexFilePath = "{}.{}".format(
                outputFilePath, AlignmentFileConstants.BAI.lower())
            utils.log("Creating index file '{}'".format(indexFilePath))
            pysam.index(outputFilePath)



"""
Tests that the project's module import graph conforms to certain policies
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import fnmatch
import itertools
import logging
import operator
import os
import pprint
import unittest
import sys

from snakefood.util import iter_pyfiles, setup_logging, is_python
from snakefood.find import find_dependencies
from snakefood.find import ERROR_IMPORT, ERROR_SYMBOL, ERROR_UNUSED
from snakefood.fallback.collections import defaultdict
from snakefood.roots import find_roots, relfile

import tests.paths as paths


class TestImports(unittest.TestCase):
    """
    Tests that the import graph:
    - doesn't contain any cycles
    - doesn't violate layering constraints
    """
    @classmethod
    def setUpClass(cls):
        snakefoodScanner = SnakefoodScanner()
        cls.graph = snakefoodScanner.scan()

    def testNoCycles(self):
        checker = ImportGraphCycleChecker(self.graph)
        checker.checkNoCycles()

    def testLayering(self):
        checker = ImportGraphLayerChecker(self.graph)
        checker.checkLayeringEnforced()


##############
# Exceptions #
##############


class ConfigurationException(Exception):
    """
    The configuration of a policy checker is invalid
    """


class PolicyException(Exception):
    """
    The code violates some enforced policy
    """


class SnakefoodScannerException(Exception):
    """
    Something went wrong in the snakefood wrapper
    """


###############
# ImportGraph #
###############


class ImportGraphNodeColor(object):
    """
    Node color constants for cycle detection
    """
    WHITE = "WHITE"  # unvisited
    BLACK = "BLACK"  # visited
    GREY = "GREY"  # currently visiting


class ImportGraphNode(object):
    """
    A node in the import graph
    """
    def __init__(self, entry):
        self.name = entry.from_filename
        self.deps = set([entry.to_filename])
        # below fields are for cycle detection
        self.color = ImportGraphNodeColor.WHITE

    def __repr__(self):
        return "ImportGraphNode: {} -> {}".format(
            self.name, repr(list(self.deps)))


class ImportGraph(object):
    """
    A directed graph of import relationships.
    Nodes are files/modules and edges are dependencies.
    """
    def __init__(self):
        self.graph = {}

    def iterNodes(self):
        return self.graph.items()

    def getNodeFor(self, name):
        return self.graph[name]

    def addEntry(self, entry):
        if entry.to_filename is None:
            return
        if entry.from_filename in self.graph:
            self.graph[entry.from_filename].deps.add(entry.to_filename)
        else:
            node = ImportGraphNode(entry)
            self.graph[entry.from_filename] = node

    def hasAnyDependencies(self, name):
        return name in self.graph and len(self.graph[name].deps) != 0

    def hasDependencyOn(self, name, dependency):
        if name not in self.graph:
            return False
        return dependency in self.graph[name].deps

    def removeDependency(self, name, dependency):
        self.graph[name].deps.remove(dependency)

    def printGraph(self):
        pprint.pprint(self.graph)


############
# Checkers #
############


class ImportGraphLayerChecker(object):
    """
    Checks the import graph layering policy

    TODO this class could be more efficient with some optimizations,
    but as it stands the time for these tests is dominated by file
    operations and and parsing the ASTs
    """
    excludedPythonFilenames = set(['__init__.py', '_version.py'])

    # each file/module is in one and only one moduleGroup
    moduleGroupNames = {
        'cli': [
            'ga4gh/server/cli/client.py',
            'ga4gh/server/cli/common.py',
            'ga4gh/server/cli/configtest.py',
            'ga4gh/server/cli/repomanager.py',
            'ga4gh/server/cli/server.py',
        ],
        'client': [
            'ga4gh/server/client.py',
        ],
        'frontend': [
            'ga4gh/server/frontend.py',
            'ga4gh/server/repo_manager.py',
        ],
        'backend': [
            'ga4gh/server/backend.py',
            'ga4gh/server/datarepo.py',
            'ga4gh/server/paging.py',
            'ga4gh/server/response_builder.py',
        ],
        'exceptions': [
            'ga4gh/server/exceptions.py',
        ],
        'datamodel': [
            'ga4gh/server/datamodel/bio_metadata.py',
            'ga4gh/server/datamodel/reads.py',
            'ga4gh/server/datamodel/references.py',
            'ga4gh/server/datamodel/rna_quantification.py',
            'ga4gh/server/datamodel/variants.py',
            'ga4gh/server/datamodel/datasets.py',
            'ga4gh/server/datamodel/ontologies.py',
            'ga4gh/server/datamodel/obo_parser.py',
            'ga4gh/server/datamodel/sequence_annotations.py',
            'ga4gh/server/datamodel/continuous.py',
            'ga4gh/server/datamodel/genotype_phenotype.py',
            'ga4gh/server/datamodel/genotype_phenotype_featureset.py',
            'ga4gh/server/datamodel/peers.py',
            'ga4gh/server/gff3.py',
            'ga4gh/server/sqlite_backend.py',
        ],
        'libraries': [
            'ga4gh/server/converters.py',
            'ga4gh/server/configtest.py',
        ],
        'config': [
            'ga4gh/server/serverconfig.py',
        ],
        'repo': [
            'ga4gh/server/repo/rnaseq2ga.py',
            'ga4gh/server/repo/models.py',
        ],
    }

    # each moduleGroupName has one and only one entry here
    layers = [
        ['cli'],
        ['client'],
        ['frontend'],
        ['backend'],
        ['libraries'],
        ['datamodel'],
        ['repo'],
        ['exceptions'],
        ['config'],
    ]

    def __init__(self, graph):
        self._checkConfiguration()
        self.graph = graph
        self.moduleGroupToOrderIndex = {}
        for i, layerRow in enumerate(self.layers):
            for moduleGroup in layerRow:
                self.moduleGroupToOrderIndex[moduleGroup] = i
        self.moduleToModuleGroup = {}
        for moduleGroup, modules in self.moduleGroupNames.items():
            for module in modules:
                self.moduleToModuleGroup[module] = moduleGroup

    def checkLayeringEnforced(self):
        # rules:
        # - no module can import from modules in layers above it
        # - no module can import from modules in moduleGroups in
        #       same layer as it
        for layer in self.layers:
            for moduleGroup in layer:
                modulesInGroup = self.moduleGroupNames[moduleGroup]
                self._sameLayerCheck(layer, moduleGroup, modulesInGroup)
                self._aboveLayerCheck(layer, moduleGroup, modulesInGroup)

    def _allModules(self):
        modules = list(itertools.chain(*self.moduleGroupNames.values()))
        return modules

    def _checkConfiguration(self):
        # each module that exists in the file tree appears in moduleGroupNames
        pythonFiles = []
        for root, dirnames, filenames in os.walk(paths.getGa4ghFilePath()):
            for filename in fnmatch.filter(filenames, '*.py'):
                pythonFilename = os.path.relpath(
                    os.path.join(root, filename))
                if (pythonFilename not in self.excludedPythonFilenames and
                        filename not in self.excludedPythonFilenames):
                    pythonFiles.append(pythonFilename)
        modules = self._allModules()
        moduleSet = set(modules)
        for pythonFile in pythonFiles:
            if pythonFile not in moduleSet:
                message = "file {} is not listed in moduleGroupNames".format(
                    pythonFile)
                raise ConfigurationException(message)

        # each module should only appear once in moduleGroupNames
        modules = self._allModules()
        moduleSet = set(modules)
        if len(modules) != len(moduleSet):
            for module in moduleSet:
                modules.remove(module)
            message = "duplicate module names in moduleGroupNames: {}"
            raise ConfigurationException(message.format(', '.join(modules)))

        # each moduleGroup should only appear once in layers
        # every defined moduleGroup appears in layers
        moduleGroups = self.moduleGroupNames.keys()
        layersModuleGroups = list(itertools.chain(*self.layers))
        if set(moduleGroups) != set(layersModuleGroups):
            message = "moduleGroupNames and layer moduleGroups not equal"
            raise ConfigurationException(message)

    def _layerIndex(self, layerName):
        return self.moduleGroupToOrderIndex[layerName]

    def _moduleGroupNamesAtSameLayerAs(self, moduleGroup):
        layerIndex = self._layerIndex(moduleGroup)
        layerCopy = self.layers[layerIndex][::]
        layerCopy.remove(moduleGroup)
        return layerCopy

    def _modulesInModuleGroup(self, moduleGroup):
        return self.moduleGroupNames[moduleGroup]

    def _modulesAtSameLayerAs(self, moduleGroup):
        moduleGroupNamesAtSameLayer = self._moduleGroupNamesAtSameLayerAs(
            moduleGroup)
        modules = []
        for moduleGroupName in moduleGroupNamesAtSameLayer:
            layerModules = self._modulesInModuleGroup(moduleGroupName)
            modules.extend(layerModules)
        return modules

    def _modulesAtLayerIndex(self, layerIndex):
        modules = []
        for moduleGroup in self.layers[layerIndex]:
            modules.extend(self._modulesInModuleGroup(moduleGroup))
        return modules

    def _modulesInLayersAbove(self, moduleGroup):
        layerIndex = self._layerIndex(moduleGroup)
        layersAbove = self.layers[:layerIndex]
        modules = []
        for i, layer in enumerate(layersAbove):
            layerModules = self._modulesAtLayerIndex(i)
            modules.extend(layerModules)
        return modules

    def _sameLayerCheck(self, layer, moduleGroup, modulesInGroup):
        modulesAtSameLayer = self._modulesAtSameLayerAs(moduleGroup)
        for module in modulesInGroup:
            for sameLayerModule in modulesAtSameLayer:
                if self.graph.hasDependencyOn(module, sameLayerModule):
                    message = "module '{}' in moduleGroup '{}' " \
                        "has dependency on module '{}' in same layer '{}'"
                    exceptionString = message.format(
                        module, moduleGroup, sameLayerModule, layer)
                    raise PolicyException(exceptionString)

    def _aboveLayerCheck(self, layer, moduleGroup, modulesInGroup):
        modulesAboveLayer = self._modulesInLayersAbove(moduleGroup)
        for module in modulesInGroup:
            for aboveLayerModule in modulesAboveLayer:
                if self.graph.hasDependencyOn(module, aboveLayerModule):
                    group = self.moduleToModuleGroup[aboveLayerModule]
                    message = "module '{}' in moduleGroup '{}' " \
                        "has dependency on module '{}' in moduleGroup '{}'"
                    exceptionString = message.format(
                        module, moduleGroup,
                        aboveLayerModule, group)
                    raise PolicyException(exceptionString)


class ImportGraphCycleChecker(object):
    """
    Checks that there are no cycles in the import graph
    (except those that are explicitly allowed)
    """

    # cyclic dependencies that we want to exclude from validation;
    # essentially, an entry here removes an edge from the dependency
    # graph as far as cycle detection is concerned
    cycleExclusions = [
    ]

    def __init__(self, graph):
        self.graph = graph
        self.visitStack = []

    def checkNoCycles(self):
        graph = self._getPreprocessedGraph()
        for name, node in graph.iterNodes():
            if node.color == ImportGraphNodeColor.WHITE:
                self._visitNode(graph, node)

    def _getPreprocessedGraph(self):
        graph = copy.deepcopy(self.graph)
        for name, dependency in self.cycleExclusions:
            graph.removeDependency(name, dependency)
        return graph

    def _visitNode(self, graph, node):
        self.visitStack.append(node)
        node.color = ImportGraphNodeColor.GREY
        for dependency in node.deps:
            if not graph.hasAnyDependencies(dependency):
                continue
            dependencyNode = graph.getNodeFor(dependency)
            if dependencyNode.color == ImportGraphNodeColor.GREY:
                self.visitStack.append(dependencyNode)
                pathString = ' --> '.join(
                    [visited.name for visited in self.visitStack])
                exceptionStr = "Circular import reference: {}".format(
                    pathString)
                raise PolicyException(exceptionStr)
            elif dependencyNode.color == ImportGraphNodeColor.WHITE:
                self._visitNode(graph, dependencyNode)
        node.color = ImportGraphNodeColor.BLACK
        self.visitStack.pop()


#############
# Snakefood #
#############


class SnakefoodEntries(object):
    """
    A list of import entries that snakefood generates
    """
    def __init__(self):
        self.entries = []

    def append(self, entry):
        self.entries.append(entry)

    def printEntries(self):
        pprint.pprint(self.entries)

    def iterEntries(self):
        return iter(self.entries)


class SnakefoodEntry(object):
    """
    An import record that snakefood generates
    """
    def __init__(self, from_root, from_filename, to_root, to_filename):
        self.from_root = from_root
        self.from_filename = from_filename
        self.to_root = to_root
        self.to_filename = to_filename

    def __repr__(self):
        return "SnakefoodEntry: {} -> {}".format(
            self.from_filename, self.to_filename)


class SnakefoodScanner(object):
    """
    Scans for imports within modules in the project.
    Mostly taken from here:
    https://bitbucket.org/blais/snakefood/src/
    e0a74fa6260dcd44716d40b4eb404ca024323eac/
    lib/python/snakefood/gendeps.py?at=default
    """
    def __init__(self):
        self.optsIgnoreUnused = None
        self.optsVerbose = 0
        self.optsDoPragmas = True
        self.optsQuiet = 1
        self.optsInternal = 1
        self.optsExternal = None
        self.optsIgnores = ['.svn', 'CVS', 'build', '.hg', '.git']
        self.optsPrintRoots = None
        self.optsFollow = True
        self.args = [paths.packageName]

    def scan(self):
        """
        Returns an ImportGraph
        """
        self.optsVerbose -= self.optsQuiet
        setup_logging(self.optsVerbose)
        info = logging.info
        warning = logging.warning
        debug = logging.debug
        if self.optsInternal and self.optsExternal:
            message = "Using --internal and --external at the same time " \
                "does not make sense."
            raise SnakefoodScannerException(message)
        if self.optsPrintRoots:
            inroots = find_roots(self.args, self.optsIgnores)
            for dn in sorted(inroots):
                print(dn)
            return
        info("")
        info("Input paths:")
        for arg in self.args:
            fn = os.path.realpath(arg)
            info('  {}'.format(fn))
            if not os.path.exists(fn):
                message = "Filename '{}' does not exist.".format(fn)
                raise SnakefoodScannerException(message)
        # Get the list of package roots for our input files and prepend
        # them to the module search path to insure localized imports.
        inroots = find_roots(self.args, self.optsIgnores)
        if (self.optsInternal or self.optsExternal) and not inroots:
            message = "No package roots found from the given files or " \
                "directories. Using --internal with these roots will  " \
                "generate no dependencies."
            raise SnakefoodScannerException(message)
        info("")
        info("Roots of the input files:")
        for root in inroots:
            info('  {}'.format(root))
        info("")
        info("Using the following import path to search for modules:")
        sys.path = inroots + sys.path
        for dn in sys.path:
            info("  {}".format(dn))
        inroots = frozenset(inroots)
        # Find all the dependencies.
        info("")
        info("Processing files:")
        info("")
        allfiles = defaultdict(set)
        allerrors = []
        processed_files = set()
        fiter = iter_pyfiles(self.args, self.optsIgnores, False)
        while 1:
            newfiles = set()
            for fn in fiter:
                if fn in processed_files:
                    continue  # Make sure we process each file only once.
                info("  {}".format(fn))
                processed_files.add(fn)
                if is_python(fn):
                    files, errors = find_dependencies(
                        fn, self.optsVerbose,
                        self.optsDoPragmas, self.optsVerbose)
                    allerrors.extend(errors)
                else:
                    # If the file is not a source file, we don't know how
                    # to get the dependencies of that (without importing,
                    # which we want to avoid).
                    files = []
                # When packages are the source of dependencies, remove the
                # __init__ file.  This is important because the targets
                # also do not include the __init__ (i.e. when "from
                # <package> import <subpackage>" is seen).
                if os.path.basename(fn) == '__init__.py':
                    fn = os.path.dirname(fn)
                # Make sure all the files at least appear in the output,
                # even if it has no dependency.
                from_ = relfile(fn, self.optsIgnores)
                if from_ is None:
                    continue
                infrom = from_[0] in inroots
                if self.optsInternal and not infrom:
                    continue
                if not self.optsExternal:
                    allfiles[from_].add((None, None))
                # Add the dependencies.
                for dfn in files:
                    xfn = dfn
                    if os.path.basename(xfn) == '__init__.py':
                        xfn = os.path.dirname(xfn)
                    to_ = relfile(xfn, self.optsIgnores)
                    into = to_[0] in inroots
                    if (self.optsInternal and not into) or \
                            (self.optsExternal and into):
                        continue
                    allfiles[from_].add(to_)
                    newfiles.add(dfn)
            if not (self.optsFollow and newfiles):
                break
            else:
                fiter = iter(sorted(newfiles))
        # If internal is used twice, we filter down  further the
        # dependencies to the set of files that were processed only,
        # not just to the files that live in the same roots.
        if self.optsInternal >= 2:
            filtfiles = type(allfiles)()
            for from_, tolist in allfiles.iteritems():
                filtfiles[from_] = set(
                    x for x in tolist if x in allfiles or x == (None, None))
            allfiles = filtfiles
        info("")
        info("SUMMARY")
        info("=======")
        # Output a list of the symbols that could not
        # be imported as modules.
        reports = [
            ("Modules that were ignored because not used:",
                ERROR_UNUSED, info),
            ("Modules that could not be imported:",
                ERROR_IMPORT, warning),
            ]
        if self.optsVerbose >= 2:
            reports.append(
                ("Symbols that could not be imported as modules:",
                    ERROR_SYMBOL, debug))
        for msg, errtype, efun in reports:
            names = set(name for (err, name) in allerrors if err is errtype)
            if names:
                efun("")
                efun(msg)
                for name in sorted(names):
                    efun("  {}".format(name))
        # Output the list of roots found.
        info("")
        info("Found roots:")
        foundRoots = set()
        for key, files in allfiles.iteritems():
            foundRoots.add(key[0])
            foundRoots.update(map(operator.itemgetter(0), files))
        if None in foundRoots:
            foundRoots.remove(None)
        for root in sorted(foundRoots):
            info("  {}".format(root))
        # Output the dependencies.
        entries = SnakefoodEntries()
        info("")
        for (from_root, from_), targets in sorted(
                allfiles.iteritems(), key=operator.itemgetter(0)):
            for to_root, to_ in sorted(targets):
                entry = SnakefoodEntry(from_root, from_, to_root, to_)
                entries.append(entry)
        graph = ImportGraph()
        for entry in entries.iterEntries():
            graph.addEntry(entry)
        return graph



"""
Module responsible for translating read data into GA4GH native
objects.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
import json
import os.path
import random

import pysam

import ga4gh.server.datamodel as datamodel
import ga4gh.server.datamodel.references as references
import ga4gh.server.exceptions as exceptions

import ga4gh.schemas.pb as pb
import ga4gh.schemas.protocol as protocol


def parseMalformedBamHeader(headerDict):
    """
    Parses the (probably) intended values out of the specified
    BAM header dictionary, which is incompletely parsed by pysam.
    This is caused by some tools incorrectly using spaces instead
    of tabs as a seperator.
    """
    headerString = " ".join(
        "{}:{}".format(k, v) for k, v in headerDict.items() if k != 'CL')
    ret = {}
    for item in headerString.split():
        key, value = item.split(":", 1)
        # build up dict, casting everything back to original type
        ret[key] = type(headerDict.get(key, ""))(value)
    if 'CL' in headerDict:
        ret['CL'] = headerDict['CL']
    return ret


class SamCigar(object):
    """
    Utility class for working with SAM CIGAR strings
    """
    # see http://pysam.readthedocs.org/en/latest/api.html
    # #pysam.AlignedSegment.cigartuples
    cigarStrings = [
        protocol.CigarUnit.ALIGNMENT_MATCH,
        protocol.CigarUnit.INSERT,
        protocol.CigarUnit.DELETE,
        protocol.CigarUnit.SKIP,
        protocol.CigarUnit.CLIP_SOFT,
        protocol.CigarUnit.CLIP_HARD,
        protocol.CigarUnit.PAD,
        protocol.CigarUnit.SEQUENCE_MATCH,
        protocol.CigarUnit.SEQUENCE_MISMATCH,
    ]

    @classmethod
    def ga2int(cls, value):
        for i, cigarString in enumerate(cls.cigarStrings):
            if value == cigarString:
                return i

    @classmethod
    def int2ga(cls, value):
        return cls.cigarStrings[value]


class SamFlags(object):
    """
    Utility class for working with SAM flags
    """
    READ_PAIRED = 0x1
    READ_PROPER_PAIR = 0x2
    READ_UNMAPPED = 0x4
    MATE_UNMAPPED = 0x8
    READ_REVERSE_STRAND = 0x10
    MATE_REVERSE_STRAND = 0x20
    FIRST_IN_PAIR = 0x40
    SECOND_IN_PAIR = 0x80
    SECONDARY_ALIGNMENT = 0x100
    FAILED_QUALITY_CHECK = 0x200
    DUPLICATE_READ = 0x400
    SUPPLEMENTARY_ALIGNMENT = 0x800

    @staticmethod
    def isFlagSet(flagAttr, flag):
        return flagAttr & flag == flag

    @staticmethod
    def setFlag(flagAttr, flag):
        return flagAttr | flag


class AlignmentDataMixin(datamodel.PysamDatamodelMixin):
    """
    Mixin class that provides methods for getting read alignments
    from bam files
    """
    def _getReadAlignments(
            self, reference, start, end, readGroupSet, readGroup):
        """
        Returns an iterator over the specified reads
        """
        # TODO If reference is None, return against all references,
        # including unmapped reads.
        samFile = self.getFileHandle(self._dataUrl)
        referenceName = reference.getLocalId().encode()
        # TODO deal with errors from htslib
        start, end = self.sanitizeAlignmentFileFetch(start, end)
        readAlignments = samFile.fetch(referenceName, start, end)
        for readAlignment in readAlignments:
            tags = dict(readAlignment.tags)
            if readGroup is None:
                if 'RG' in tags:
                    alignmentReadGroupLocalId = tags['RG']
                    readGroupCompoundId = datamodel.ReadGroupCompoundId(
                        readGroupSet.getCompoundId(),
                        str(alignmentReadGroupLocalId))
                yield self.convertReadAlignment(
                    readAlignment, readGroupSet, str(readGroupCompoundId))
            else:
                if self._filterReads:
                    if 'RG' in tags and tags['RG'] == self._localId:
                        yield self.convertReadAlignment(
                            readAlignment, readGroupSet,
                            str(readGroup.getCompoundId()))
                else:
                    yield self.convertReadAlignment(
                        readAlignment, readGroupSet,
                        str(readGroup.getCompoundId()))

    def convertReadAlignment(self, read, readGroupSet, readGroupId):
        """
        Convert a pysam ReadAlignment to a GA4GH ReadAlignment
        """
        samFile = self.getFileHandle(self._dataUrl)
        # TODO fill out remaining fields
        # TODO refine in tandem with code in converters module
        ret = protocol.ReadAlignment()
        # ret.fragmentId = 'TODO'
        ret.aligned_quality.extend(read.query_qualities)
        ret.aligned_sequence = read.query_sequence
        if SamFlags.isFlagSet(read.flag, SamFlags.READ_UNMAPPED):
            ret.ClearField("alignment")
        else:
            ret.alignment.CopyFrom(protocol.LinearAlignment())
            ret.alignment.mapping_quality = read.mapping_quality
            ret.alignment.position.CopyFrom(protocol.Position())
            ret.alignment.position.reference_name = samFile.getrname(
                read.reference_id)
            ret.alignment.position.position = read.reference_start
            ret.alignment.position.strand = protocol.POS_STRAND
            if SamFlags.isFlagSet(read.flag, SamFlags.READ_REVERSE_STRAND):
                ret.alignment.position.strand = protocol.NEG_STRAND
            for operation, length in read.cigar:
                gaCigarUnit = ret.alignment.cigar.add()
                gaCigarUnit.operation = SamCigar.int2ga(operation)
                gaCigarUnit.operation_length = length
                gaCigarUnit.reference_sequence = ""  # TODO fix this!
        ret.duplicate_fragment = SamFlags.isFlagSet(
            read.flag, SamFlags.DUPLICATE_READ)
        ret.failed_vendor_quality_checks = SamFlags.isFlagSet(
            read.flag, SamFlags.FAILED_QUALITY_CHECK)
        ret.fragment_length = read.template_length
        ret.fragment_name = read.query_name
        for key, value in read.tags:
            # Useful for inspecting the structure of read tags
            # print("{key} {ktype}: {value}, {vtype}".format(
            #     key=key, ktype=type(key), value=value, vtype=type(value)))
            protocol.setAttribute(ret.attributes.attr[key].values, value)

        if SamFlags.isFlagSet(read.flag, SamFlags.MATE_UNMAPPED):
            ret.next_mate_position.Clear()
        else:
            ret.next_mate_position.Clear()
            if read.next_reference_id != -1:
                ret.next_mate_position.reference_name = samFile.getrname(
                    read.next_reference_id)
            else:
                ret.next_mate_position.reference_name = ""
            ret.next_mate_position.position = read.next_reference_start
            ret.next_mate_position.strand = protocol.POS_STRAND
            if SamFlags.isFlagSet(read.flag, SamFlags.MATE_REVERSE_STRAND):
                ret.next_mate_position.strand = protocol.NEG_STRAND
        if SamFlags.isFlagSet(read.flag, SamFlags.READ_PAIRED):
            ret.number_reads = 2
        else:
            ret.number_reads = 1
        ret.read_number = -1
        if SamFlags.isFlagSet(read.flag, SamFlags.FIRST_IN_PAIR):
            if SamFlags.isFlagSet(read.flag, SamFlags.SECOND_IN_PAIR):
                ret.read_number = 2
            else:
                ret.read_number = 0
        elif SamFlags.isFlagSet(read.flag, SamFlags.SECOND_IN_PAIR):
            ret.read_number = 1
        ret.improper_placement = not SamFlags.isFlagSet(
            read.flag, SamFlags.READ_PROPER_PAIR)
        ret.read_group_id = readGroupId
        ret.secondary_alignment = SamFlags.isFlagSet(
            read.flag, SamFlags.SECONDARY_ALIGNMENT)
        ret.supplementary_alignment = SamFlags.isFlagSet(
            read.flag, SamFlags.SUPPLEMENTARY_ALIGNMENT)
        ret.id = readGroupSet.getReadAlignmentId(ret)
        return ret

    def openFile(self, dataFile):
        # We need to check to see if the path exists here as pysam does
        # not throw an error if the index is missing.
        if not os.path.exists(self._indexFile):
            raise exceptions.FileOpenFailedException(self._indexFile)
        try:
            return pysam.AlignmentFile(
                self._dataUrl, filepath_index=self._indexFile)
        except IOError as exception:
            # IOError thrown when the index file passed in is not actually
            # an index file... may also happen in other cases?
            raise exceptions.DataException(exception.message)


class AbstractReadGroupSet(datamodel.DatamodelObject):
    """
    The base class of a read group set
    """
    compoundIdClass = datamodel.ReadGroupSetCompoundId

    def __init__(self, parentContainer, localId):
        super(AbstractReadGroupSet, self).__init__(parentContainer, localId)
        self._readGroupIdMap = {}
        self._readGroupIds = []
        self._referenceSet = None
        self._numAlignedReads = -1
        self._numUnalignedReads = -1

    def setReferenceSet(self, referenceSet):
        """
        Sets the reference set for this ReadGroupSet to the specified value.
        """
        self._referenceSet = referenceSet

    def addReadGroup(self, readGroup):
        """
        Adds the specified ReadGroup to this ReadGroupSet.
        """
        id_ = readGroup.getId()
        self._readGroupIdMap[id_] = readGroup
        self._readGroupIds.append(id_)

    def getReadGroups(self):
        """
        Returns the list of ReadGroups in this ReadGroupSet.
        """
        return [self._readGroupIdMap[id_] for id_ in self._readGroupIds]

    def getReadGroupIds(self):
        """
        Returns the list of readGroupIds in this ReadGroupSet.
        """
        return self._readGroupIds

    def getReadGroup(self, id_):
        """
        Returns the ReadGroup with the specified id if it exists in this
        ReadGroupSet, or raises a ReadGroupNotFoundException otherwise.
        """
        if id_ not in self._readGroupIdMap:
            raise exceptions.ReadGroupNotFoundException(id_)
        return self._readGroupIdMap[id_]

    def getReferenceSet(self):
        """
        Returns the ReferenceSet that this ReadGroupSet is aligned to.
        """
        return self._referenceSet

    def toProtocolElement(self):
        """
        Returns the GA4GH protocol representation of this ReadGroupSet.
        """
        readGroupSet = protocol.ReadGroupSet()
        readGroupSet.id = self.getId()
        readGroupSet.read_groups.extend(
            [readGroup.toProtocolElement()
             for readGroup in self.getReadGroups()]
        )
        readGroupSet.name = self.getLocalId()
        readGroupSet.dataset_id = self.getParentContainer().getId()
        readGroupSet.stats.CopyFrom(self.getStats())
        self.serializeAttributes(readGroupSet)
        return readGroupSet

    def getNumAlignedReads(self):
        """
        Return the number of aligned reads in this read group set
        """
        return self._numAlignedReads

    def getNumUnalignedReads(self):
        """
        Return the number of unaligned reads in this read group set
        """
        return self._numUnalignedReads

    def getPrograms(self):
        """
        Returns an array of Programs used to generate this read group set
        """
        raise NotImplementedError()

    def getReadAlignmentId(self, gaAlignment):
        """
        Returns a string ID suitable for use in the specified GA
        ReadAlignment object in this ReadGroupSet.
        """
        compoundId = datamodel.ReadAlignmentCompoundId(
            self.getCompoundId(), gaAlignment.fragment_name)
        return str(compoundId)

    def getStats(self):
        """
        Returns the GA4GH protocol representation of this read group set's
        ReadStats.
        """
        stats = protocol.ReadStats()
        stats.aligned_read_count = self._numAlignedReads
        stats.unaligned_read_count = self._numUnalignedReads
        return stats


class SimulatedReadGroupSet(AbstractReadGroupSet):
    """
    A simulated read group set
    """
    def __init__(
            self, parentContainer, localId, referenceSet, randomSeed=1,
            numReadGroups=1, numAlignments=2):
        super(SimulatedReadGroupSet, self).__init__(
            parentContainer, localId)
        self._referenceSet = referenceSet
        self._numAlignments = numAlignments
        self._numAlignedReads = self._numAlignments
        self._numUnalignedReads = 0
        for i in range(numReadGroups):
            localId = "rg{}".format(i)
            readGroup = SimulatedReadGroup(
                self, localId, randomSeed + i, numAlignments)
            self.addReadGroup(readGroup)

    def getPrograms(self):
        return []

    def getReadAlignments(self, referenceId=None, start=None, end=None):
        for readGroup in self.getReadGroups():
            iterator = readGroup.getReadAlignments(referenceId, start, end)
            for alignment in iterator:
                yield alignment


class HtslibReadGroupSet(AlignmentDataMixin, AbstractReadGroupSet):
    """
    Class representing a logical collection ReadGroups.
    """
    defaultReadGroupName = "default"

    def __init__(self, parentContainer, localId):
        super(HtslibReadGroupSet, self).__init__(parentContainer, localId)
        self._programs = []
        self._dataUrl = None
        self._indexFile = None
        # Used when we populate from a file. Not defined when we populate
        # from the DB.
        self._bamHeaderReferenceSetName = None

    def getReadAlignments(self, reference, start=None, end=None):
        """
        Returns an iterator over the specified reads
        """
        return self._getReadAlignments(reference, start, end, self, None)

    def getBamHeaderReferenceSetName(self):
        """
        Returns the ReferenceSet name using in the BAM header.
        """
        return self._bamHeaderReferenceSetName

    def populateFromRow(self, readGroupSetRecord):
        """
        Populates the instance variables of this ReadGroupSet from the
        specified database row.
        """
        self._dataUrl = readGroupSetRecord.dataurl
        self._indexFile = readGroupSetRecord.indexfile
        self._programs = []
        for jsonDict in json.loads(readGroupSetRecord.programs):
            program = protocol.fromJson(json.dumps(jsonDict),
                                        protocol.Program)
            self._programs.append(program)
        stats = protocol.fromJson(readGroupSetRecord.stats, protocol.ReadStats)
        self._numAlignedReads = stats.aligned_read_count
        self._numUnalignedReads = stats.unaligned_read_count

    def populateFromFile(self, dataUrl, indexFile=None):
        """
        Populates the instance variables of this ReadGroupSet from the
        specified dataUrl and indexFile. If indexFile is not specified
        guess usual form.
        """
        self._dataUrl = dataUrl
        self._indexFile = indexFile
        if indexFile is None:
            self._indexFile = dataUrl + ".bai"
        samFile = self.getFileHandle(self._dataUrl)
        self._setHeaderFields(samFile)
        if 'RG' not in samFile.header or len(samFile.header['RG']) == 0:
            readGroup = HtslibReadGroup(self, self.defaultReadGroupName)
            self.addReadGroup(readGroup)
        else:
            for readGroupHeader in samFile.header['RG']:
                readGroup = HtslibReadGroup(self, readGroupHeader['ID'])
                readGroup.populateFromHeader(readGroupHeader)
                self.addReadGroup(readGroup)
        self._bamHeaderReferenceSetName = None
        for referenceInfo in samFile.header['SQ']:
            if 'AS' not in referenceInfo:
                infoDict = parseMalformedBamHeader(referenceInfo)
            else:
                infoDict = referenceInfo
            name = infoDict.get('AS', references.DEFAULT_REFERENCESET_NAME)
            if self._bamHeaderReferenceSetName is None:
                self._bamHeaderReferenceSetName = name
            elif self._bamHeaderReferenceSetName != name:
                raise exceptions.MultipleReferenceSetsInReadGroupSet(
                    self._dataUrl, name, self._bamFileReferenceName)
        self._numAlignedReads = samFile.mapped
        self._numUnalignedReads = samFile.unmapped

    def checkConsistency(self, dataRepository):
        pass
        # TODO verify that the references in the BAM file exist
        # in the reference set. Otherwise, we won't be able to
        # query for them.

    def _setHeaderFields(self, samFile):
        programs = []
        if 'PG' in samFile.header:
            htslibPrograms = samFile.header['PG']
            for htslibProgram in htslibPrograms:
                program = protocol.Program()
                program.id = htslibProgram['ID']
                program.command_line = htslibProgram.get(
                    'CL', pb.DEFAULT_STRING)
                program.name = htslibProgram.get('PN', pb.DEFAULT_STRING)
                program.prev_program_id = htslibProgram.get(
                    'PP', pb.DEFAULT_STRING)
                program.version = htslibProgram.get('VN', pb.DEFAULT_STRING)
                programs.append(program)
        self._programs = programs

    def getPrograms(self):
        return self._programs

    def getDataUrl(self):
        """
        Returns the data URL for this ReadGroupSet.
        """
        return self._dataUrl

    def getIndexFile(self):
        """
        Returns the index file for this ReadGroupSet.
        """
        return self._indexFile


class AbstractReadGroup(datamodel.DatamodelObject):
    """
    Class representing a ReadGroup. A ReadGroup is all the data that's
    processed the same way by the sequencer.  There are typically 1-10
    ReadGroups in a ReadGroupSet.
    """
    compoundIdClass = datamodel.ReadGroupCompoundId

    def __init__(self, parentContainer, localId):
        super(AbstractReadGroup, self).__init__(parentContainer, localId)
        datetimeNow = datetime.datetime.now()
        now = protocol.convertDatetime(datetimeNow)
        self._iso8601 = datetimeNow.strftime("%Y-%m-%dT%H:%M:%SZ")
        self._creationTime = now
        self._updateTime = now
        self._biosampleId = None

    def toProtocolElement(self):
        """
        Returns the GA4GH protocol representation of this ReadGroup.
        """
        # TODO this is very incomplete, but we don't have the
        # implementation to fill out the rest of the fields currently
        readGroup = protocol.ReadGroup()
        readGroup.id = self.getId()
        readGroup.created = self._creationTime
        readGroup.updated = self._updateTime
        dataset = self.getParentContainer().getParentContainer()
        readGroup.dataset_id = dataset.getId()
        readGroup.name = self.getLocalId()
        readGroup.predicted_insert_size = pb.int(self.getPredictedInsertSize())
        referenceSet = self._parentContainer.getReferenceSet()
        readGroup.sample_name = pb.string(self.getSampleName())
        readGroup.biosample_id = pb.string(self.getBiosampleId())
        if referenceSet is not None:
            readGroup.reference_set_id = referenceSet.getId()
        readGroup.stats.CopyFrom(self.getStats())
        readGroup.programs.extend(self.getPrograms())
        readGroup.description = pb.string(self.getDescription())
        readGroup.experiment.CopyFrom(self.getExperiment())
        self.serializeAttributes(readGroup)
        return readGroup

    def getStats(self):
        """
        Returns the GA4GH protocol representation of this read group's
        ReadStats.
        """
        stats = protocol.ReadStats()
        stats.aligned_read_count = self.getNumAlignedReads()
        stats.unaligned_read_count = self.getNumUnalignedReads()
        # TODO base_count requires iterating through all reads
        return stats

    def getExperiment(self):
        """
        Returns the GA4GH protocol representation of this read group's
        Experiment.
        """
        experiment = protocol.Experiment()
        experiment.id = self.getExperimentId()
        experiment.instrument_model = pb.string(self.getInstrumentModel())
        experiment.sequencing_center = pb.string(self.getSequencingCenter())
        experiment.description = pb.string(self.getExperimentDescription())
        experiment.library = pb.string(self.getLibrary())
        experiment.platform_unit = pb.string(self.getPlatformUnit())
        experiment.message_create_time = self._iso8601
        experiment.message_update_time = self._iso8601
        experiment.run_time = pb.string(self.getRunTime())
        return experiment

    def getNumAlignedReads(self):
        """
        Return the number of aligned reads in the read group
        """
        return self._numAlignedReads

    def getNumUnalignedReads(self):
        """
        Return the number of unaligned reads in the read group
        """
        return self._numUnalignedReads

    def getPrograms(self):
        """
        Returns an array of Programs used to generate this read group
        """
        raise NotImplementedError()

    def getBiosampleId(self):
        return self._biosampleId

    def setBiosampleId(self, biosampleId):
        self._biosampleId = biosampleId

    def getDescription(self):
        """
        Returns a description of this read group
        """
        raise NotImplementedError()

    def getSampleName(self):
        """
        Returns the sample id of the read group
        """
        raise NotImplementedError()

    def getPredictedInsertSize(self):
        """
        Returns the predicted insert size of the read group
        """
        raise NotImplementedError()

    def getInstrumentModel(self):
        """
        Returns the instrument model used for this experiment
        """
        raise NotImplementedError()

    def getSequencingCenter(self):
        """
        Returns the sequencing center used for this experiment
        """
        raise NotImplementedError()

    def getExperimentDescription(self):
        """
        Returns the description of this read group
        """
        raise NotImplementedError()

    def getLibrary(self):
        """
        Returns the name of the library used in this experiment
        """
        raise NotImplementedError()

    def getPlatformUnit(self):
        """
        Returns the platform unit used in this experiment
        """
        raise NotImplementedError()

    def getRunTime(self):
        """
        Returns the time at which the experiment was performed
        """
        raise NotImplementedError()

    def getExperimentId(self):
        """
        Returns the id of the experiment used for this read group
        """
        return str(datamodel.ExperimentCompoundId(
            self.getCompoundId(), 'experiment'))


class SimulatedReadGroup(AbstractReadGroup):
    """
    A simulated readgroup
    """

    def __init__(self, parentContainer, localId, randomSeed, numAlignments=2):
        super(SimulatedReadGroup, self).__init__(parentContainer, localId)
        self._randomSeed = randomSeed
        self._numAlignedReads = self._parentContainer.getNumAlignedReads()
        self._numUnalignedReads = 0

    def getReadAlignments(self, referenceId=None, start=None, end=None):
        rng = random.Random(self._randomSeed)

        # We seed reads with sequential seeds starting from here. We hope no
        # ranges for two different simulated read groups ever overlap (because
        # then we'd start seeing identical reads in the two groups.)
        read_seed_start = rng.getrandbits(64)

        for i in range(self.getNumAlignedReads()):
            seed = read_seed_start + i
            yield self._createReadAlignment(i, seed)

    def _createReadAlignment(self, i, seed):
        # TODO fill out a bit more
        rng = random.Random(seed)
        alignment = protocol.ReadAlignment()
        alignment.fragment_length = rng.randint(10, 100)
        alignment.aligned_sequence = ""
        for i in range(alignment.fragment_length):
            # TODO: are these reasonable quality values?
            alignment.aligned_quality.append(rng.randint(1, 20))
            alignment.aligned_sequence += rng.choice("ACGT")

        alignment.alignment.position.position = 0
        alignment.alignment.position.reference_name = "NotImplemented"
        alignment.alignment.position.strand = protocol.POS_STRAND
        alignment.duplicate_fragment = False
        alignment.failed_vendor_quality_checks = False

        alignment.fragment_name = "{}$simulated{}".format(
            self.getLocalId(), i)
        alignment.number_reads = 0
        alignment.improper_placement = False
        alignment.read_group_id = self.getId()
        alignment.read_number = -1
        alignment.secondary_alignment = False
        alignment.supplementary_alignment = False
        alignment.id = self._parentContainer.getReadAlignmentId(alignment)
        return alignment

    def getPrograms(self):
        return []

    def getDescription(self):
        return None

    def getSampleName(self):
        return 'sampleId'

    def getPredictedInsertSize(self):
        return 0

    def getInstrumentModel(self):
        return None

    def getSequencingCenter(self):
        return None

    def getExperimentDescription(self):
        return None

    def getLibrary(self):
        return None

    def getPlatformUnit(self):
        return None

    def getRunTime(self):
        return None


class HtslibReadGroup(AlignmentDataMixin, AbstractReadGroup):
    """
    A readgroup based on htslib's reading of a given file
    """
    def __init__(self, parentContainer, localId):
        super(HtslibReadGroup, self).__init__(parentContainer, localId)
        # These attributes are used in AlignmentDataMixin.openFile
        self._dataUrl = parentContainer.getDataUrl()
        self._indexFile = parentContainer.getIndexFile()
        self._filterReads = localId != HtslibReadGroupSet.defaultReadGroupName
        self._biosampleId = None
        self._sampleName = None
        self._description = None
        self._predictedInsertSize = None
        self._instrumentModel = None
        self._sequencingCenter = None
        self._experimentDescription = None
        self._library = None
        self._platformUnit = None
        self._runTime = None
        self._numAlignedReads = -1  # TODO populate with metadata
        self._numUnalignedReads = -1  # TODO populate with metadata

    def populateFromHeader(self, readGroupHeader):
        """
        Populate the instance variables using the specified SAM header.
        """
        self._sampleName = readGroupHeader.get('SM', None)
        self._description = readGroupHeader.get('DS', None)
        if 'PI' in readGroupHeader:
            self._predictedInsertSize = int(readGroupHeader['PI'])
        self._instrumentModel = readGroupHeader.get('PL', None)
        self._sequencingCenter = readGroupHeader.get('CN', None)
        self._experimentDescription = readGroupHeader.get('DS', None)
        self._library = readGroupHeader.get('LB', None)
        self._platformUnit = readGroupHeader.get('PU', None)
        self._runTime = readGroupHeader.get('DT', None)

    def populateFromRow(self, readGroupRecord):
        """
        Populate the instance variables using the specified DB row.
        """
        self._sampleName = readGroupRecord.samplename
        self._biosampleId = readGroupRecord.biosampleid
        self._description = readGroupRecord.description
        self._predictedInsertSize = readGroupRecord.predictedinsertsize
        stats = protocol.fromJson(readGroupRecord.stats, protocol.ReadStats)
        self._numAlignedReads = stats.aligned_read_count
        self._numUnalignedReads = stats.unaligned_read_count
        experiment = protocol.fromJson(
            readGroupRecord.experiment, protocol.Experiment)
        self._instrumentModel = experiment.instrument_model
        self._sequencingCenter = experiment.sequencing_center
        self._experimentDescription = experiment.description
        self._library = experiment.library
        self._platformUnit = experiment.platform_unit
        self._runTime = experiment.run_time

    def getReadAlignments(self, reference, start=None, end=None):
        """
        Returns an iterator over the specified reads
        """
        return self._getReadAlignments(
            reference, start, end, self._parentContainer, self)

    def getPrograms(self):
        return self._parentContainer.getPrograms()

    def getDescription(self):
        return self._description

    def getSampleName(self):
        return self._sampleName

    def getPredictedInsertSize(self):
        return self._predictedInsertSize

    def getInstrumentModel(self):
        return self._instrumentModel

    def getSequencingCenter(self):
        return self._sequencingCenter

    def getExperimentDescription(self):
        return self._experimentDescription

    def getLibrary(self):
        return self._library

    def getPlatformUnit(self):
        return self._platformUnit

    def getRunTime(self):
        return self._runTime



"""
End to end test that invokes the repo manager
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import mock
import os
import tempfile
import unittest
import json

import ga4gh.server.cli.repomanager as cli_repomanager
import tests.paths as paths

import ga4gh.schemas.protocol as protocol


class RepoManagerEndToEndTest(unittest.TestCase):

    datasetName = 'datasetOne'
    metadata = {'description': 'aDescription'}
    individualName = "test"
    biosampleName = "test"
    individual = protocol.toJson(protocol.Individual(
        name="test",
        created="2016-05-19T21:00:19Z",
        updated="2016-05-19T21:00:19Z"))
    biosample = protocol.toJson(protocol.Biosample(
        name="test",
        created="2016-05-19T21:00:19Z",
        updated="2016-05-19T21:00:19Z"))
    attributes = {"key1": "value1", "key2": "value2"}

    def setUp(self):
        _, self.repoFile = tempfile.mkstemp(
            prefix='ga4gh_repo_manager_end2end_test')
        os.unlink(self.repoFile)

    def tearDown(self):
        if os.path.exists(self.repoFile):
            os.unlink(self.repoFile)

    def _runCmd(self, cmd, *args):
        command = [cmd, self.repoFile] + list(args)
        cli_repomanager.repo_main(command)

    def testEndToEnd(self):
        self._runCmd("init")
        self._runCmd("list-announcements")
        self._runCmd("clear-announcements")
        self._runCmd("add-peer", paths.peerUrl)
        self._runCmd("add-ontology", paths.ontologyPath)
        self._runCmd(
            "add-referenceset", paths.faPath,
            '-n', paths.referenceSetName,
            '-A', json.dumps(self.attributes))
        self._runCmd("add-dataset", self.datasetName,
                     '-A', json.dumps(self.attributes))
        self._runCmd("add-biosample",
                     self.datasetName,
                     self.biosampleName,
                     self.biosample)
        self._runCmd("add-individual",
                     self.datasetName,
                     self.individualName,
                     self.individual)
        self._runCmd(
            "add-readgroupset", self.datasetName, paths.bamPath,
            '-R', paths.referenceSetName, '-n', paths.readGroupSetName,
            '-A', json.dumps(self.attributes))
        self._runCmd(
            "add-featureset", self.datasetName, paths.featuresPath,
            '-R', paths.referenceSetName, '-O', paths.ontologyName,
            '-A', json.dumps(self.attributes))
        # ensure we can handle trailing slashes
        vcfPath = paths.vcfDirPath + '/'
        self._runCmd(
            "add-variantset", self.datasetName,
            vcfPath, '-R', paths.referenceSetName,
            '-A', json.dumps(self.attributes))
        variantAnnotationSetName = "vas"
        self._runCmd(
            "add-variantset", self.datasetName,
            paths.annotatedVcfPath, '-R', paths.referenceSetName,
            "-aO", paths.ontologyName, "-n", variantAnnotationSetName,
            '-A', json.dumps(self.attributes))
        phenotypeAssociationSetName = "paSet"
        self._runCmd(
            "add-phenotypeassociationset",
            self.datasetName,
            paths.phenotypeAssociationSetPath,
            "-n",
            phenotypeAssociationSetName,
            '-A', json.dumps(self.attributes))

        self._runCmd("verify")
        self._runCmd("list")
        self._runCmd("remove-peer", paths.peerUrl, "-f")
        self._runCmd(
            "remove-phenotypeassociationset",
            self.datasetName, phenotypeAssociationSetName, "-f")
        self._runCmd(
            "remove-variantset", self.datasetName, paths.variantSetName, "-f")
        self._runCmd(
            "remove-variantset", self.datasetName, variantAnnotationSetName,
            "-f")
        self._runCmd(
            "remove-readgroupset", self.datasetName,
            paths.readGroupSetName, "-f")
        self._runCmd(
            "remove-featureset", self.datasetName, paths.featureSetName,
            "-f")
        self._runCmd(
            "remove-dataset", self.datasetName, "-f")
        self._runCmd(
            "remove-referenceset", paths.referenceSetName, "-f")
        self._runCmd(
            "remove-ontology", paths.ontologyName, "-f")

    def testForce(self):
        datasetName = 'dataset1'
        self._runCmd("init")
        self._runCmd("add-dataset", datasetName)
        with mock.patch(
                'ga4gh.server.cli.repomanager.getRawInput',
                lambda x: 'N'):
            self._runCmd("remove-dataset", datasetName)
        with mock.patch(
                'ga4gh.server.cli.repomanager.getRawInput',
                lambda x: 'y'):
            self._runCmd("remove-dataset", datasetName)
            with self.assertRaises(SystemExit):
                self._runCmd("remove-dataset", datasetName)



"""
Data-driven tests for reads
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections
import os

import ga4gh.server.backend as backend
import ga4gh.server.datamodel as datamodel
import ga4gh.server.datamodel.datasets as datasets
import ga4gh.server.datamodel.reads as reads
import ga4gh.server.datamodel.references as references
import ga4gh.server.datarepo as datarepo
import tests.datadriven as datadriven
import tests.paths as paths

import ga4gh.common.utils as utils
import ga4gh.schemas.protocol as protocol

import pysam


def testReads():
    testDataDir = os.path.join(paths.testDataDir, "datasets/dataset1/reads")
    for test in datadriven.makeTests(
            testDataDir, ReadGroupSetTest, '*.bam'):
        yield test


class ReadGroupSetInfo(object):
    """
    Container class for information about a read group set
    """
    def __init__(self, samFile):
        self.numAlignedReads = samFile.mapped
        self.numUnalignedReads = samFile.unmapped


class ReadGroupInfo(object):
    """
    Container class for information about a read group
    """
    def __init__(self, gaReadGroupSet, samFile, readGroupName):
        self.gaReadGroup = reads.AbstractReadGroup(
            gaReadGroupSet, readGroupName)
        self.id = self.gaReadGroup.getId()
        self.samFile = samFile
        self.mappedReads = collections.defaultdict(list)
        for read in self.samFile:
            tags = dict(read.tags)
            if 'RG' not in tags or tags['RG'] != readGroupName:
                continue
            if read.reference_id != -1:
                # mapped read
                referenceName = self.samFile.getrname(read.reference_id)
                self.mappedReads[referenceName].append(read)
        self.numAlignedReads = -1
        self.numUnalignedReads = -1
        self.programs = []
        if 'PG' in self.samFile.header:
            self.programs = self.samFile.header['PG']
        self.sampleName = None
        self.description = None
        self.predictedInsertSize = None
        self.instrumentModel = None
        self.sequencingCenter = None
        self.experimentDescription = None
        self.library = None
        self.platformUnit = None
        self.runTime = None
        if 'RG' in self.samFile.header:
            readGroupHeader = [
                rgHeader for rgHeader in self.samFile.header['RG']
                if rgHeader['ID'] == readGroupName][0]
            self.sampleName = readGroupHeader.get('SM', None)
            self.description = readGroupHeader.get('DS', None)
            if 'PI' in readGroupHeader:
                self.predictedInsertSize = int(readGroupHeader['PI'])
            self.instrumentModel = readGroupHeader.get('PL', None)
            self.sequencingCenter = readGroupHeader.get('CN', None)
            self.experimentDescription = readGroupHeader.get('DS', None)
            self.library = readGroupHeader.get('LB', None)
            self.platformUnit = readGroupHeader.get('PU', None)
            self.runTime = readGroupHeader.get('DT', None)


class ReadGroupSetTest(datadriven.DataDrivenTest):
    """
    Data driven test for read group sets
    """
    def __init__(self, localId, dataPath):
        self._backend = backend.Backend(datarepo.AbstractDataRepository())
        self._referenceSet = None
        self._dataset = datasets.Dataset("ds")
        self._readGroupInfos = {}
        self._readGroupSetInfo = None
        self._samFile = pysam.AlignmentFile(dataPath)
        self._readReferences()
        super(ReadGroupSetTest, self).__init__(localId, dataPath)
        self._readAlignmentInfo()

    def _readReferences(self):
        # Read the reference information from the samfile
        referenceSetName = None
        for referenceInfo in self._samFile.header['SQ']:
            if 'AS' not in referenceInfo:
                infoDict = reads.parseMalformedBamHeader(referenceInfo)
            # If there's still no reference set name in there we use
            # a default name.
            name = infoDict.get("AS", "Default")
            if referenceSetName is None:
                referenceSetName = name
                self._addReferenceSet(referenceSetName)
            else:
                self.assertEqual(referenceSetName, name)
            self._addReference(infoDict['SN'])

    def _addReferenceSet(self, referenceSetName):
        self._referenceSet = references.AbstractReferenceSet(referenceSetName)
        self._backend.getDataRepository().addReferenceSet(self._referenceSet)

    def _addReference(self, referenceName):
        reference = references.AbstractReference(
            self._referenceSet, referenceName)
        self._referenceSet.addReference(reference)

    def _readAlignmentInfo(self):
        self._readGroupSetInfo = ReadGroupSetInfo(self._samFile)
        if 'RG' in self._samFile.header:
            readGroupHeaders = self._samFile.header['RG']
            readGroupNames = [
                readGroupHeader['ID'] for readGroupHeader
                in readGroupHeaders]
        else:
            readGroupNames = ['default']
        for readGroupName in readGroupNames:
            readGroupInfo = ReadGroupInfo(
                self._gaObject, self._samFile, readGroupName)
            self._readGroupInfos[readGroupName] = readGroupInfo

    def getDataModelInstance(self, localId, dataPath):
        readGroupSet = reads.HtslibReadGroupSet(self._dataset, localId)
        readGroupSet.populateFromFile(dataPath)
        return readGroupSet

    def getProtocolClass(self):
        return protocol.ReadGroupSet

    def testSampleNameEtc(self):
        # test that sampleId and other misc fields are set correctly
        readGroupSet = self._gaObject
        for readGroup in readGroupSet.getReadGroups():
            readGroupInfo = self._readGroupInfos[readGroup.getLocalId()]
            gaReadGroup = readGroup.toProtocolElement()
            self.assertEqual(
                readGroupInfo.sampleName,
                gaReadGroup.sample_name)
            self.assertEqual(
                readGroupInfo.predictedInsertSize,
                gaReadGroup.predicted_insert_size)
            self.assertEqual(
                readGroupInfo.description,
                gaReadGroup.description)

    def testExperiments(self):
        # test that the experiment field is set correctly
        readGroupSet = self._gaObject
        for readGroup in readGroupSet.getReadGroups():
            readGroupInfo = self._readGroupInfos[readGroup.getLocalId()]
            gaReadGroup = readGroup.toProtocolElement()
            self.assertIn(
                "experiment",
                datamodel.CompoundId.deobfuscate(gaReadGroup.experiment.id))
            self.assertEqual(
                readGroupInfo.instrumentModel,
                gaReadGroup.experiment.instrument_model)
            self.assertEqual(
                readGroupInfo.sequencingCenter,
                gaReadGroup.experiment.sequencing_center)
            self.assertEqual(
                readGroupInfo.experimentDescription,
                gaReadGroup.experiment.description)
            self.assertEqual(
                readGroupInfo.library,
                gaReadGroup.experiment.library)
            self.assertEqual(
                readGroupInfo.platformUnit,
                gaReadGroup.experiment.platform_unit)
            self.assertEqual(
                readGroupInfo.runTime,
                gaReadGroup.experiment.run_time)

    def testPrograms(self):
        # test that program info is set correctly
        readGroupSet = self._gaObject
        for readGroup in readGroupSet.getReadGroups():
            readGroupInfo = self._readGroupInfos[readGroup.getLocalId()]
            gaPrograms = readGroup.getPrograms()
            htslibPrograms = readGroupInfo.programs
            for gaProgram, htslibProgram in utils.zipLists(
                    gaPrograms, htslibPrograms):
                self.assertEqual(
                    gaProgram.id, htslibProgram.get('ID'))
                self.assertEqual(
                    gaProgram.command_line, htslibProgram.get('CL', None))
                self.assertEqual(
                    gaProgram.name, htslibProgram.get('PN', None))
                self.assertEqual(
                    gaProgram.prev_program_id, htslibProgram.get('PP', None))
                self.assertEqual(
                    gaProgram.version, htslibProgram.get('VN', None))

    def testReadGroupStats(self):
        # test that the stats attrs are populated correctly
        readGroupSet = self._gaObject
        gaReadGroupSet = readGroupSet.toProtocolElement()
        readGroupSetInfo = self._readGroupSetInfo
        self.assertEqual(
            readGroupSet.getNumAlignedReads(),
            readGroupSetInfo.numAlignedReads)
        self.assertEqual(
            readGroupSet.getNumUnalignedReads(),
            readGroupSetInfo.numUnalignedReads)
        self.assertEqual(
            gaReadGroupSet.stats.aligned_read_count,
            readGroupSetInfo.numAlignedReads)
        self.assertEqual(
            gaReadGroupSet.stats.unaligned_read_count,
            readGroupSetInfo.numUnalignedReads)
        for readGroup in readGroupSet.getReadGroups():
            gaReadGroup = readGroup.toProtocolElement()
            self.assertEqual(
                readGroup.getNumAlignedReads(), -1)
            self.assertEqual(
                readGroup.getNumUnalignedReads(), -1)
            self.assertEqual(
                gaReadGroup.stats.aligned_read_count, -1)
            self.assertEqual(
                gaReadGroup.stats.unaligned_read_count, -1)

    def testValidateObjects(self):
        # test that validation works on read groups and reads
        readGroupSet = self._gaObject
        for readGroup in readGroupSet.getReadGroups():
            self.assertIsInstance(
                readGroup.toProtocolElement(), protocol.ReadGroup)
            for reference in self._referenceSet.getReferences():
                for gaAlignment in readGroup.getReadAlignments(reference):
                    self.assertIsInstance(
                        gaAlignment, protocol.ReadAlignment)

    def testGetReadAlignmentsRefId(self):
        # test that searching with a reference id succeeds
        readGroupSet = self._gaObject
        for readGroup in readGroupSet.getReadGroups():
            readGroupInfo = self._readGroupInfos[readGroup.getLocalId()]
            for name, alignments in readGroupInfo.mappedReads.items():
                reference = self._referenceSet.getReferenceByName(name)
                self.assertAlignmentListsEqual(
                    list(readGroup.getReadAlignments(reference)), alignments,
                    readGroupInfo)

    def testGetReadAlignmentsStartEnd(self):
        # test that searching with start and end coords succeeds
        readGroupSet = self._gaObject
        for readGroup in readGroupSet.getReadGroups():
            readGroupInfo = self._readGroupInfos[readGroup.getLocalId()]
            for name, alignments, in readGroupInfo.mappedReads.items():
                bigNumThatPysamWontChokeOn = 2**30
                reference = self._referenceSet.getReferenceByName(name)
                gaAlignments = list(readGroup.getReadAlignments(
                    reference, 0, bigNumThatPysamWontChokeOn))
                self.assertAlignmentListsEqual(
                    gaAlignments, alignments, readGroupInfo)

    def testGetReadAlignmentSearchRanges(self):
        # test that various range searches work
        readGroupSet = self._gaObject
        for readGroup in readGroupSet.getReadGroups():
            readGroupInfo = self._readGroupInfos[readGroup.getLocalId()]
            for name in readGroupInfo.mappedReads.keys():
                reference = self._referenceSet.getReferenceByName(name)
                alignments = list(readGroup.getReadAlignments(reference))
                length = len(alignments)
                if length < 2:
                    continue
                positions = [
                    read.alignment.position.position for read in alignments
                    if read.alignment is not None]
                if length != len(set(positions)):
                    continue
                begin = positions[0]
                end = positions[-1]
                self.assertGetReadAlignmentsRangeResult(
                    readGroup, reference, begin, end + 1, length)
                self.assertGetReadAlignmentsRangeResult(
                    readGroup, reference, begin, end, length - 1)
                self.assertGetReadAlignmentsRangeResult(
                    readGroup, reference, begin, begin, 0)

    def assertGetReadAlignmentsRangeResult(
            self, readGroup, reference, start, end, result):
        alignments = list(readGroup.getReadAlignments(reference, start, end))
        self.assertEqual(len(alignments), result)

    def assertAlignmentListsEqual(
            self, gaAlignments, pysamAlignments, readGroupInfo):
        for gaAlignment, pysamAlignment in utils.zipLists(
                gaAlignments, pysamAlignments):
            self.assertAlignmentsEqual(
                gaAlignment, pysamAlignment, readGroupInfo)

    def getDictFromMessageMap(self, messageMap):
        return dict([
            (k, [protocol.getValueFromValue(x) for x in v.values])
            for (k, v) in messageMap._values.items()])

    def assertAlignmentsEqual(self, gaAlignment, pysamAlignment,
                              readGroupInfo):
        if pysamAlignment.query_qualities is None:
            self.assertEqual(gaAlignment.aligned_quality, [])
        else:
            self.assertEqual(
                gaAlignment.aligned_quality,
                list(pysamAlignment.query_qualities))
        self.assertEqual(
            gaAlignment.aligned_sequence,
            pysamAlignment.query_sequence)
        if reads.SamFlags.isFlagSet(
                pysamAlignment.flag, reads.SamFlags.READ_UNMAPPED):
            self.assertEqual(0, gaAlignment.alignment.ByteSize())
        else:
            self.assertEqual(
                gaAlignment.alignment.mapping_quality,
                pysamAlignment.mapping_quality)
            self.assertEqual(
                gaAlignment.alignment.position.reference_name,
                readGroupInfo.samFile.getrname(pysamAlignment.reference_id))
            self.assertEqual(
                gaAlignment.alignment.position.position,
                pysamAlignment.reference_start)
            # TODO test reverseStrand on position and on
            # nextMatePosition once it has been implemented.
            self.assertCigarEqual(
                gaAlignment.alignment.cigar,
                pysamAlignment.cigar)
        self.assertFlag(
            gaAlignment.duplicate_fragment,
            pysamAlignment, reads.SamFlags.DUPLICATE_READ)
        self.assertFlag(
            gaAlignment.failed_vendor_quality_checks,
            pysamAlignment, reads.SamFlags.FAILED_QUALITY_CHECK)
        self.assertEqual(
            gaAlignment.fragment_length,
            pysamAlignment.template_length)
        self.assertEqual(
            gaAlignment.fragment_name,
            pysamAlignment.query_name)
        compoundId = datamodel.ReadAlignmentCompoundId(
            self._gaObject.getCompoundId(),
            pysamAlignment.query_name)
        self.assertEqual(gaAlignment.id, str(compoundId))
        ret = protocol.ReadAlignment()
        for key, value in pysamAlignment.tags:
            protocol.setAttribute(ret.attributes.attr[key].values, value)
        self.assertEqual(
            protocol.toJson(gaAlignment.attributes),
            protocol.toJson(ret.attributes))
        if reads.SamFlags.isFlagSet(
                pysamAlignment.flag, reads.SamFlags.MATE_UNMAPPED):
            self.assertEqual(0, gaAlignment.next_mate_position.ByteSize())
        else:
            self.assertEqual(
                gaAlignment.next_mate_position.position,
                pysamAlignment.next_reference_start)
            if pysamAlignment.next_reference_id != -1:
                self.assertEqual(
                    gaAlignment.next_mate_position.reference_name,
                    readGroupInfo.samFile.getrname(
                        pysamAlignment.next_reference_id))
            else:
                self.assertEqual(
                    gaAlignment.next_mate_position.reference_name, "")
        if gaAlignment.number_reads == 1:
            self.assertFlag(
                False, pysamAlignment, reads.SamFlags.READ_PAIRED)
        elif gaAlignment.number_reads == 2:
            self.assertFlag(
                True, pysamAlignment, reads.SamFlags.READ_PAIRED)
        else:
            # we shouldn't be setting numberReads to anything else
            self.assertTrue(False)
        if gaAlignment.read_number is -1:
            self.assertFlag(
                False, pysamAlignment, reads.SamFlags.FIRST_IN_PAIR)
            self.assertFlag(
                False, pysamAlignment, reads.SamFlags.SECOND_IN_PAIR)
        elif gaAlignment.read_number == 0:
            self.assertFlag(
                True, pysamAlignment, reads.SamFlags.FIRST_IN_PAIR)
            self.assertFlag(
                False, pysamAlignment, reads.SamFlags.SECOND_IN_PAIR)
        elif gaAlignment.read_number == 1:
            self.assertFlag(
                False, pysamAlignment, reads.SamFlags.FIRST_IN_PAIR)
            self.assertFlag(
                True, pysamAlignment, reads.SamFlags.SECOND_IN_PAIR)
        elif gaAlignment.read_number == 2:
            self.assertFlag(
                True, pysamAlignment, reads.SamFlags.FIRST_IN_PAIR)
            self.assertFlag(
                True, pysamAlignment, reads.SamFlags.SECOND_IN_PAIR)
        else:
            # we shouldn't be setting readNumber to anything else
            self.assertTrue(False)
        self.assertFlag(
            not gaAlignment.improper_placement,
            pysamAlignment, reads.SamFlags.READ_PROPER_PAIR)
        self.assertEqual(
            gaAlignment.read_group_id,
            readGroupInfo.id)
        self.assertFlag(
            gaAlignment.secondary_alignment,
            pysamAlignment, reads.SamFlags.SECONDARY_ALIGNMENT)
        self.assertFlag(
            gaAlignment.supplementary_alignment,
            pysamAlignment, reads.SamFlags.SUPPLEMENTARY_ALIGNMENT)

    def assertFlag(self, gaAlignmentAttr, pysamAlignment, mask):
        flagSet = reads.SamFlags.isFlagSet(pysamAlignment.flag, mask)
        self.assertEqual(gaAlignmentAttr, flagSet)

    def assertCigarEqual(self, gaCigar, pysamCigar):
        self.assertEqual(len(gaCigar), len(pysamCigar))
        for i, gaCigarUnit in enumerate(gaCigar):
            operation, length = pysamCigar[i]
            gaCigarUnitOperation = reads.SamCigar.ga2int(
                gaCigarUnit.operation)
            self.assertEqual(
                gaCigarUnitOperation, operation)
            self.assertEqual(
                gaCigarUnit.operation_length, length)
